import json
from labelbox.client import Client
from labelbox.data.annotation_types.classification.classification import Checklist, ClassificationAnnotation, ClassificationAnswer, Radio
from labelbox.data.annotation_types.data.video import VideoData
from labelbox.data.annotation_types.geometry.line import Line
from labelbox.data.annotation_types.geometry.point import Point
from labelbox.data.annotation_types.geometry.rectangle import Rectangle
from labelbox.data.annotation_types.geometry.point import Point

from labelbox.data.annotation_types.label import Label
from labelbox.data.annotation_types.video import VideoObjectAnnotation
from labelbox import parser

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_video():
    with open('tests/data/assets/ndjson/video_import.json', 'r') as file:
        data = json.load(file)

    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == [data[2], data[0], data[1], data[3], data[4], data[5]]


def test_video_name_only():
    with open('tests/data/assets/ndjson/video_import_name_only.json',
              'r') as file:
        data = json.load(file)

    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == [data[2], data[0], data[1], data[3], data[4], data[5]]


def test_video_classification_global_subclassifications():
    label = Label(
        data=VideoData(global_key="sample-video-4.mp4",),
        annotations=[
            ClassificationAnnotation(
                name='radio_question_nested',
                value=Radio(answer=ClassificationAnswer(
                    name='first_radio_question')),
            ),
            ClassificationAnnotation(
                name='nested_checklist_question',
                value=Checklist(
                    name='checklist',
                    answer=[
                        ClassificationAnswer(
                            name='first_checklist_answer',
                            classifications=[
                                ClassificationAnnotation(
                                    name='sub_checklist_question',
                                    value=Radio(answer=ClassificationAnswer(
                                        name='first_sub_checklist_answer')))
                            ])
                    ]))
        ])

    expected_first_annotation = {
        'name': 'radio_question_nested',
        'answer': {
            'name': 'first_radio_question'
        },
        'dataRow': {
            'globalKey': 'sample-video-4.mp4'
        }
    }

    expected_second_annotation = nested_checklist_annotation_ndjson = {
        "name": "nested_checklist_question",
        "answer": [{
            "name":
                "first_checklist_answer",
            "classifications": [{
                "name": "sub_checklist_question",
                "answer": {
                    "name": "first_sub_checklist_answer"
                }
            }]
        }],
        'dataRow': {
            'globalKey': 'sample-video-4.mp4'
        }
    }

    serialized = NDJsonConverter.serialize([label])
    res = [x for x in serialized]
    for annotations in res:
        annotations.pop("uuid")
    assert res == [expected_first_annotation, expected_second_annotation]

    deserialized = NDJsonConverter.deserialize(res)
    res = next(deserialized)
    annotations = res.annotations
    for annotation in annotations:
        annotation.extra.pop("uuid")
    assert annotations == label.annotations


def test_video_classification_nesting_bbox():
    bbox_annotation = [
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=13,
            segment_index=0,
            value=Rectangle(
                start=Point(x=146.0, y=98.0),  # Top left
                end=Point(x=382.0, y=341.0),  # Bottom right
            ),
            classifications=[
                ClassificationAnnotation(
                    name='radio_question_nested',
                    value=Radio(answer=ClassificationAnswer(
                        name='first_radio_question',
                        classifications=[
                            ClassificationAnnotation(name='sub_question_radio',
                                                     value=Checklist(answer=[
                                                         ClassificationAnswer(
                                                             name='sub_answer')
                                                     ]))
                        ])),
                )
            ]),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=15,
            segment_index=0,
            value=Rectangle(
                start=Point(x=146.0, y=98.0),  # Top left
                end=Point(x=382.0, y=341.0),  # Bottom right,
            ),
            classifications=[
                ClassificationAnnotation(
                    name='nested_checklist_question',
                    value=Checklist(answer=[
                        ClassificationAnswer(
                            name='first_checklist_answer',
                            classifications=[
                                ClassificationAnnotation(
                                    name='sub_checklist_question',
                                    value=Radio(answer=ClassificationAnswer(
                                        name='first_sub_checklist_answer')))
                            ])
                    ]))
            ]),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=19,
            segment_index=0,
            value=Rectangle(
                start=Point(x=146.0, y=98.0),  # Top left
                end=Point(x=382.0, y=341.0),  # Bottom right
            ))
    ]
    expected = [{
        'dataRow': {
            'globalKey': 'sample-video-4.mp4'
        },
        'name':
            'bbox_video',
        'classifications': [],
        'segments': [{
            'keyframes': [{
                'frame':
                    13,
                'bbox': {
                    'top': 98.0,
                    'left': 146.0,
                    'height': 243.0,
                    'width': 236.0
                },
                'classifications': [{
                    'name': 'radio_question_nested',
                    'answer': {
                        'name':
                            'first_radio_question',
                        'classifications': [{
                            'name': 'sub_question_radio',
                            'answer': [{
                                'name': 'sub_answer'
                            }]
                        }]
                    }
                }]
            }, {
                'frame':
                    15,
                'bbox': {
                    'top': 98.0,
                    'left': 146.0,
                    'height': 243.0,
                    'width': 236.0
                },
                'classifications': [{
                    'name':
                        'nested_checklist_question',
                    'answer': [{
                        'name':
                            'first_checklist_answer',
                        'classifications': [{
                            'name': 'sub_checklist_question',
                            'answer': {
                                'name': 'first_sub_checklist_answer'
                            }
                        }]
                    }]
                }]
            }, {
                'frame': 19,
                'bbox': {
                    'top': 98.0,
                    'left': 146.0,
                    'height': 243.0,
                    'width': 236.0
                },
                'classifications': []
            }]
        }]
    }]

    label = Label(data=VideoData(global_key="sample-video-4.mp4",),
                  annotations=bbox_annotation)

    serialized = NDJsonConverter.serialize([label])
    res = [x for x in serialized]
    for annotations in res:
        annotations.pop("uuid")
    assert res == expected

    deserialized = NDJsonConverter.deserialize(res)
    res = next(deserialized)
    annotations = res.annotations
    for annotation in annotations:
        annotation.extra.pop("uuid")
    assert annotations == label.annotations


def test_video_classification_point():
    bbox_annotation = [
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=13,
            segment_index=0,
            value=Point(x=46.0, y=8.0),
            classifications=[
                ClassificationAnnotation(
                    name='radio_question_nested',
                    value=Radio(answer=ClassificationAnswer(
                        name='first_radio_question',
                        classifications=[
                            ClassificationAnnotation(name='sub_question_radio',
                                                     value=Checklist(answer=[
                                                         ClassificationAnswer(
                                                             name='sub_answer')
                                                     ]))
                        ])),
                )
            ]),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=15,
            segment_index=0,
            value=Point(x=56.0, y=18.0),
            classifications=[
                ClassificationAnnotation(
                    name='nested_checklist_question',
                    value=Checklist(answer=[
                        ClassificationAnswer(
                            name='first_checklist_answer',
                            classifications=[
                                ClassificationAnnotation(
                                    name='sub_checklist_question',
                                    value=Radio(answer=ClassificationAnswer(
                                        name='first_sub_checklist_answer')))
                            ])
                    ]))
            ]),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=19,
            segment_index=0,
            value=Point(x=66.0, y=28.0),
        )
    ]
    expected = [{
        'dataRow': {
            'globalKey': 'sample-video-4.mp4'
        },
        'name':
            'bbox_video',
        'classifications': [],
        'segments': [{
            'keyframes': [{
                'frame':
                    13,
                'point': {
                    'x': 46.0,
                    'y': 8.0,
                },
                'classifications': [{
                    'name': 'radio_question_nested',
                    'answer': {
                        'name':
                            'first_radio_question',
                        'classifications': [{
                            'name': 'sub_question_radio',
                            'answer': [{
                                'name': 'sub_answer'
                            }]
                        }]
                    }
                }]
            }, {
                'frame':
                    15,
                'point': {
                    'x': 56.0,
                    'y': 18.0,
                },
                'classifications': [{
                    'name':
                        'nested_checklist_question',
                    'answer': [{
                        'name':
                            'first_checklist_answer',
                        'classifications': [{
                            'name': 'sub_checklist_question',
                            'answer': {
                                'name': 'first_sub_checklist_answer'
                            }
                        }]
                    }]
                }]
            }, {
                'frame': 19,
                'point': {
                    'x': 66.0,
                    'y': 28.0,
                },
                'classifications': []
            }]
        }]
    }]

    label = Label(data=VideoData(global_key="sample-video-4.mp4",),
                  annotations=bbox_annotation)

    serialized = NDJsonConverter.serialize([label])
    res = [x for x in serialized]
    for annotations in res:
        annotations.pop("uuid")
    assert res == expected

    deserialized = NDJsonConverter.deserialize(res)
    res = next(deserialized)
    annotations = res.annotations
    for annotation in annotations:
        annotation.extra.pop("uuid")
    assert annotations == label.annotations


def test_video_classification_frameline():
    bbox_annotation = [
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=13,
            segment_index=0,
            value=Line(
                points=[Point(x=8, y=10), Point(x=10, y=9)]),
            classifications=[
                ClassificationAnnotation(
                    name='radio_question_nested',
                    value=Radio(answer=ClassificationAnswer(
                        name='first_radio_question',
                        classifications=[
                            ClassificationAnnotation(name='sub_question_radio',
                                                     value=Checklist(answer=[
                                                         ClassificationAnswer(
                                                             name='sub_answer')
                                                     ]))
                        ])),
                )
            ]),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=15,
            segment_index=0,
            value=Line(
                points=[Point(x=18, y=20), Point(x=20, y=19)]),
            classifications=[
                ClassificationAnnotation(
                    name='nested_checklist_question',
                    value=Checklist(answer=[
                        ClassificationAnswer(
                            name='first_checklist_answer',
                            classifications=[
                                ClassificationAnnotation(
                                    name='sub_checklist_question',
                                    value=Radio(answer=ClassificationAnswer(
                                        name='first_sub_checklist_answer')))
                            ])
                    ]))
            ]),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=19,
            segment_index=0,
            value=Line(
                points=[Point(x=28, y=30), Point(x=30, y=29)]),
        )
    ]
    expected = [{
        'dataRow': {
            'globalKey': 'sample-video-4.mp4'
        },
        'name':
            'bbox_video',
        'classifications': [],
        'segments': [{
            'keyframes': [{
                'frame':
                    13,
                'line': [{
                    'x': 8.0,
                    'y': 10.0,
                }, {
                    'x': 10.0,
                    'y': 9.0,
                }],
                'classifications': [{
                    'name': 'radio_question_nested',
                    'answer': {
                        'name':
                            'first_radio_question',
                        'classifications': [{
                            'name': 'sub_question_radio',
                            'answer': [{
                                'name': 'sub_answer'
                            }]
                        }]
                    }
                }]
            }, {
                'frame':
                    15,
                'line': [{
                    'x': 18.0,
                    'y': 20.0,
                }, {
                    'x': 20.0,
                    'y': 19.0,
                }],
                'classifications': [{
                    'name':
                        'nested_checklist_question',
                    'answer': [{
                        'name':
                            'first_checklist_answer',
                        'classifications': [{
                            'name': 'sub_checklist_question',
                            'answer': {
                                'name': 'first_sub_checklist_answer'
                            }
                        }]
                    }]
                }]
            }, {
                'frame': 19,
                'line': [{
                    'x': 28.0,
                    'y': 30.0,
                }, {
                    'x': 30.0,
                    'y': 29.0,
                }],
                'classifications': []
            }]
        }]
    }]

    label = Label(data=VideoData(global_key="sample-video-4.mp4",),
                  annotations=bbox_annotation)

    serialized = NDJsonConverter.serialize([label])
    res = [x for x in serialized]
    for annotations in res:
        annotations.pop("uuid")
    assert res == expected

    deserialized = NDJsonConverter.deserialize(res)
    res = next(deserialized)
    annotations = res.annotations
    for annotation in annotations:
        annotation.extra.pop("uuid")
    assert annotations == label.annotations



import pytest

from labelbox.data.annotation_types import (Checklist, ClassificationAnswer,
                                            Dropdown, Radio, Text,
                                            ClassificationAnnotation)

from labelbox import pydantic_compat


def test_classification_answer():
    with pytest.raises(pydantic_compat.ValidationError):
        ClassificationAnswer()

    feature_schema_id = "schema_id"
    name = "my_feature"
    confidence = 0.9
    custom_metrics = [{'name': 'metric1', 'value': 2}]
    answer = ClassificationAnswer(name=name,
                                  confidence=confidence,
                                  custom_metrics=custom_metrics)

    assert answer.feature_schema_id is None
    assert answer.name == name
    assert answer.confidence == confidence
    assert answer.custom_metrics == custom_metrics

    answer = ClassificationAnswer(feature_schema_id=feature_schema_id,
                                  name=name)

    assert answer.feature_schema_id == feature_schema_id
    assert answer.name == name


def test_classification():
    answer = "1234"
    classification = ClassificationAnnotation(value=Text(answer=answer),
                                              name="a classification")
    assert classification.dict()['value']['answer'] == answer

    with pytest.raises(pydantic_compat.ValidationError):
        ClassificationAnnotation()


def test_subclass():
    answer = "1234"
    feature_schema_id = "11232"
    name = "my_feature"
    with pytest.raises(pydantic_compat.ValidationError):
        # Should have feature schema info
        classification = ClassificationAnnotation(value=Text(answer=answer))
    classification = ClassificationAnnotation(value=Text(answer=answer),
                                              name=name)
    assert classification.dict() == {
        'name': name,
        'feature_schema_id': None,
        'extra': {},
        'value': {
            'answer': answer,
        },
        'message_id': None,
    }
    classification = ClassificationAnnotation(
        value=Text(answer=answer),
        name=name,
        feature_schema_id=feature_schema_id)
    assert classification.dict() == {
        'name': None,
        'feature_schema_id': feature_schema_id,
        'extra': {},
        'value': {
            'answer': answer,
        },
        'name': name,
        'message_id': None,
    }
    classification = ClassificationAnnotation(
        value=Text(answer=answer),
        feature_schema_id=feature_schema_id,
        name=name)
    assert classification.dict() == {
        'name': name,
        'feature_schema_id': feature_schema_id,
        'extra': {},
        'value': {
            'answer': answer,
        },
        'message_id': None,
    }


def test_radio():
    answer = ClassificationAnswer(name="1",
                                  confidence=0.81,
                                  custom_metrics=[{
                                      'name': 'metric1',
                                      'value': 0.99
                                  }])
    feature_schema_id = "feature_schema_id"
    name = "my_feature"

    with pytest.raises(pydantic_compat.ValidationError):
        classification = ClassificationAnnotation(value=Radio(
            answer=answer.name))

    with pytest.raises(pydantic_compat.ValidationError):
        classification = Radio(answer=[answer])
    classification = Radio(answer=answer,)
    assert classification.dict() == {
        'answer': {
            'name': answer.name,
            'feature_schema_id': None,
            'extra': {},
            'confidence': 0.81,
            'custom_metrics': [{
                'name': 'metric1',
                'value': 0.99
            }],
        }
    }
    classification = ClassificationAnnotation(
        value=Radio(answer=answer),
        feature_schema_id=feature_schema_id,
        name=name,
        custom_metrics=[{
            'name': 'metric1',
            'value': 0.99
        }])
    assert classification.dict() == {
        'name': name,
        'feature_schema_id': feature_schema_id,
        'extra': {},
        'custom_metrics': [{
            'name': 'metric1',
            'value': 0.99
        }],
        'value': {
            'answer': {
                'name': answer.name,
                'feature_schema_id': None,
                'extra': {},
                'confidence': 0.81,
                'custom_metrics': [{
                    'name': 'metric1',
                    'value': 0.99
                }]
            },
        },
        'message_id': None,
    }


def test_checklist():
    answer = ClassificationAnswer(name="1",
                                  confidence=0.99,
                                  custom_metrics=[{
                                      'name': 'metric1',
                                      'value': 2
                                  }])
    feature_schema_id = "feature_schema_id"
    name = "my_feature"

    with pytest.raises(pydantic_compat.ValidationError):
        classification = Checklist(answer=answer.name)

    with pytest.raises(pydantic_compat.ValidationError):
        classification = Checklist(answer=answer)

    classification = Checklist(answer=[answer])
    assert classification.dict() == {
        'answer': [{
            'name': answer.name,
            'feature_schema_id': None,
            'extra': {},
            'confidence': 0.99,
            'custom_metrics': [{
                'name': 'metric1',
                'value': 2
            }],
        }]
    }
    classification = ClassificationAnnotation(
        value=Checklist(answer=[answer]),
        feature_schema_id=feature_schema_id,
        name=name,
    )
    assert classification.dict() == {
        'name': name,
        'feature_schema_id': feature_schema_id,
        'extra': {},
        'value': {
            'answer': [{
                'name': answer.name,
                'feature_schema_id': None,
                'extra': {},
                'confidence': 0.99,
                'custom_metrics': [{
                    'name': 'metric1',
                    'value': 2
                }],
            }]
        },
        'message_id': None,
    }


def test_dropdown():
    answer = ClassificationAnswer(name="1", confidence=1)
    feature_schema_id = "feature_schema_id"
    name = "my_feature"

    with pytest.raises(pydantic_compat.ValidationError):
        classification = ClassificationAnnotation(
            value=Dropdown(answer=answer.name), name="test")

    with pytest.raises(pydantic_compat.ValidationError):
        classification = Dropdown(answer=answer)
    classification = Dropdown(answer=[answer])
    assert classification.dict() == {
        'answer': [{
            'name': '1',
            'feature_schema_id': None,
            'extra': {},
            'confidence': 1
        }]
    }
    classification = ClassificationAnnotation(
        value=Dropdown(answer=[answer]),
        feature_schema_id=feature_schema_id,
        name=name)
    assert classification.dict() == {
        'name': name,
        'feature_schema_id': feature_schema_id,
        'extra': {},
        'value': {
            'answer': [{
                'name': answer.name,
                'feature_schema_id': None,
                'confidence': 1,
                'extra': {}
            }]
        },
        'message_id': None,
    }



from re import U

from labelbox.schema.data_row_metadata import _DeleteBatchDataRowMetadata
from labelbox.schema.identifiable import GlobalKey, UniqueId


def test_dict_delete_data_row_batch():
    obj = _DeleteBatchDataRowMetadata(
        data_row_identifier=UniqueId("abcd"),
        schema_ids=["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"])
    assert obj.dict() == {
        "data_row_identifier": {
            "id": "abcd",
            "id_type": "ID"
        },
        "schema_ids": [
            "clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"
        ]
    }

    obj = _DeleteBatchDataRowMetadata(
        data_row_identifier=GlobalKey("fegh"),
        schema_ids=["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"])
    assert obj.dict() == {
        "data_row_identifier": {
            "id": "fegh",
            "id_type": "GKEY"
        },
        "schema_ids": [
            "clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"
        ]
    }


def test_dict_delete_data_row_batch_by_alias():
    obj = _DeleteBatchDataRowMetadata(
        data_row_identifier=UniqueId("abcd"),
        schema_ids=["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"])
    assert obj.dict(by_alias=True) == {
        "dataRowIdentifier": {
            "id": "abcd",
            "idType": "ID"
        },
        "schemaIds": ["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"]
    }

    obj = _DeleteBatchDataRowMetadata(
        data_row_identifier=GlobalKey("fegh"),
        schema_ids=["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"])
    assert obj.dict(by_alias=True) == {
        "dataRowIdentifier": {
            "id": "fegh",
            "idType": "GKEY"
        },
        "schemaIds": ["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"]
    }



from datetime import datetime, timedelta, timezone
import json
import time

import pytest

from labelbox import LabelingFrontend
from labelbox.exceptions import InvalidQueryError, ResourceConflict


def simple_ontology():
    classifications = [{
        "name": "test_ontology",
        "instructions": "Which class is this?",
        "type": "radio",
        "options": [{
            "value": c,
            "label": c
        } for c in ["one", "two", "three"]],
        "required": True,
    }]

    return {"tools": [], "classifications": classifications}


def test_project_setup(project) -> None:
    client = project.client
    labeling_frontends = list(
        client.get_labeling_frontends(where=LabelingFrontend.name == 'Editor'))
    assert len(labeling_frontends)
    labeling_frontend = labeling_frontends[0]

    time.sleep(3)
    now = datetime.now().astimezone(timezone.utc)

    project.setup(labeling_frontend, simple_ontology())
    assert now - project.setup_complete <= timedelta(seconds=3)
    assert now - project.last_activity_time <= timedelta(seconds=3)

    assert project.labeling_frontend() == labeling_frontend
    options = list(project.labeling_frontend_options())
    assert len(options) == 1
    options = options[0]
    # TODO ensure that LabelingFrontendOptions can be obtaind by ID
    with pytest.raises(InvalidQueryError):
        assert options.labeling_frontend() == labeling_frontend
        assert options.project() == project
        assert options.organization() == client.get_organization()
    assert options.customization_options == json.dumps(simple_ontology())
    assert project.organization() == client.get_organization()
    assert project.created_by() == client.get_user()


def test_project_editor_setup(client, project, rand_gen):
    ontology_name = f"test_project_editor_setup_ontology_name-{rand_gen(str)}"
    ontology = client.create_ontology(ontology_name, simple_ontology())
    now = datetime.now().astimezone(timezone.utc)
    project.connect_ontology(ontology)
    assert now - project.setup_complete <= timedelta(seconds=3)
    assert now - project.last_activity_time <= timedelta(seconds=3)
    assert project.labeling_frontend().name == "Editor"
    assert project.organization() == client.get_organization()
    assert project.created_by() == client.get_user()
    assert project.ontology().name == ontology_name
    # Make sure that setup only creates one ontology
    time.sleep(3)  # Search takes a second
    assert [ontology.name for ontology in client.get_ontologies(ontology_name)
           ] == [ontology_name]


def test_project_connect_ontology_cant_call_multiple_times(
        client, project, rand_gen):
    ontology_name = f"test_project_editor_setup_ontology_name-{rand_gen(str)}"
    ontology = client.create_ontology(ontology_name, simple_ontology())
    project.connect_ontology(ontology)
    with pytest.raises(ValueError):
        project.connect_ontology(ontology)



from labelbox import pydantic_compat


class LabelScore(pydantic_compat.BaseModel):
    """
    A label score.

    Attributes:
        name (str)
        score (float)

    """

    name: str
    score: float



import json

import pytest

from labelbox.data.serialization.labelbox_v1.converter import LBV1Converter


def test_image():
    file_path = 'tests/data/assets/labelbox_v1/unkown_media_type_export.json'
    with open(file_path, 'r') as file:
        payload = json.load(file)

    collection = list(LBV1Converter.deserialize(payload))
    # One of the data rows is broken.
    assert len(collection) != len(payload)

    for row in payload:
        row['media_type'] = 'image'
        row['Global Key'] = None

    collection = LBV1Converter.deserialize(payload)
    for idx, serialized in enumerate(LBV1Converter.serialize(collection)):
        assert serialized.keys() == payload[idx].keys()
        for key in serialized:
            if key != 'Label':
                assert serialized[key] == payload[idx][key]
            elif key == 'Label':
                for annotation_a, annotation_b in zip(
                        serialized[key]['objects'],
                        payload[idx][key]['objects']):
                    if not len(annotation_a['classifications']):
                        # We don't add a classification key to the payload if there is no classifications.
                        annotation_a.pop('classifications')
                    annotation_b['page'] = None
                    annotation_b['unit'] = None

                    if isinstance(annotation_b.get('classifications'),
                                  list) and len(
                                      annotation_b['classifications']):
                        if isinstance(annotation_b['classifications'][0], list):
                            annotation_b['classifications'] = annotation_b[
                                'classifications'][0]

                    assert annotation_a == annotation_b


# After check the nd serializer on this shit.. It should work for almost everything (except the other horse shit..)



import pytest

from labelbox import ProjectRole
from faker import Faker

faker = Faker()

def test_org_invite(client, organization, environ, queries):
    role = client.get_roles()['LABELER']
    dummy_email = "none+{}@labelbox.com".format("".join(faker.random_letters(26)))
    invite_limit = organization.invite_limit()

    if environ.value == "prod":
        assert invite_limit.remaining > 0, "No invites available for the account associated with this key."
    elif environ.value != "staging":
        # Cannot run against local
        return

    invite = organization.invite_user(dummy_email, role)

    if environ.value == "prod":

        invite_limit_after = organization.invite_limit()
        # One user added
        assert invite_limit.remaining - invite_limit_after.remaining == 1
        # An invite shouldn't effect the user count until after it is accepted

    outstanding_invites = queries.get_invites(client)
    in_list = False

    for invite in outstanding_invites:
        if invite.uid == invite.uid:
            in_list = True
            org_role = invite.organization_role_name.lower()
            assert org_role == role.name.lower(
            ), "Role should be labeler. Found {org_role} "
    assert in_list, "Invite not found"
    queries.cancel_invite(client, invite.uid)
    assert invite_limit.remaining - organization.invite_limit().remaining == 0


def test_project_invite(client, organization, project_pack, queries):
    project_1, project_2 = project_pack
    roles = client.get_roles()
    dummy_email = "none+{}@labelbox.com".format("".join(faker.random_letters(26)))
    project_role_1 = ProjectRole(project=project_1, role=roles['LABELER'])
    project_role_2 = ProjectRole(project=project_2, role=roles['REVIEWER'])
    invite = organization.invite_user(
        dummy_email,
        roles['NONE'],
        project_roles=[project_role_1, project_role_2])

    project_invite = next(queries.get_project_invites(client, project_1.uid))

    assert set([(proj_invite.project.uid, proj_invite.role.uid)
                for proj_invite in project_invite.project_roles
               ]) == set([(proj_role.project.uid, proj_role.role.uid)
                          for proj_role in [project_role_1, project_role_2]])

    assert set([(proj_invite.project.uid, proj_invite.role.uid)
                for proj_invite in project_invite.project_roles
               ]) == set([(proj_role.project.uid, proj_role.role.uid)
                          for proj_role in [project_role_1, project_role_2]])

    project_members = project_1.members()

    project_member = [
        member for member in project_members
        if member.user().uid == client.get_user().uid
    ]

    assert len(project_member) == 1
    project_member = project_member[0]

    assert project_member.access_from == 'ORGANIZATION'
    assert project_member.role().name.upper() == roles['ADMIN'].name.upper()
    queries.cancel_invite(client, invite.uid)


@pytest.mark.skip(
    "Unable to programatically create user without accepting an email invite. Add back once there is a workaround."
)
def test_member_management(client, organization, project, project_based_user):
    roles = client.get_roles()
    assert not len(list(project_based_user.projects()))
    for role in [roles['LABELER'], roles['REVIEWER']]:

        project_based_user.upsert_project_role(project, role=role)
        members = project.members()
        is_member = False
        for member in members:
            if member.user().uid == project_based_user.uid:
                is_member = True
                assert member.role().name.upper() == role.name.upper()
                break
        assert is_member

    project_based_user.remove_from_project(project)
    is_member = False
    for member in project.members():
        assert member.user().uid != project_based_user.uid

    assert project_based_user.org_role().name.upper(
    ) == roles['NONE'].name.upper()
    for role in [
            roles['TEAM_MANAGER'], roles['ADMIN'], roles['LABELER'],
            roles['REVIEWER']
    ]:
        project_based_user.update_org_role(role)
        project_based_user.org_role().name.upper() == role.name.upper()

    organization.remove_user(project_based_user)
    for user in organization.users():
        assert project_based_user.uid != user.uid



import pytest
import uuid
from typing import Tuple

from labelbox.schema.media_type import MediaType
from labelbox import Project, Dataset
from labelbox.schema.data_row import DataRow
from labelbox.schema.label import Label

IMAGE_URL = "https://storage.googleapis.com/lb-artifacts-testing-public/sdk_integration_test/potato.jpeg"


def test_project_export_v2(client, export_v2_test_helpers,
                           configured_project_with_label,
                           wait_for_data_row_processing):
    project, dataset, data_row, label = configured_project_with_label
    data_row = wait_for_data_row_processing(client, data_row)
    label_id = label.uid

    task_name = "test_label_export_v2"
    params = {
        "include_performance_details": True,
        "include_labels": True,
        "media_type_override": MediaType.Image,
        "project_details": True,
        "data_row_details": True
    }

    task_results = export_v2_test_helpers.run_project_export_v2_task(
        project, task_name=task_name, params=params)

    for task_result in task_results:
        task_media_attributes = task_result['media_attributes']
        task_project = task_result['projects'][project.uid]
        task_project_label_ids_set = set(
            map(lambda prediction: prediction['id'], task_project['labels']))
        task_project_details = task_project['project_details']
        task_data_row = task_result['data_row']
        task_data_row_details = task_data_row['details']

        assert label_id in task_project_label_ids_set
        # data row
        assert task_data_row['id'] == data_row.uid
        assert task_data_row['external_id'] == data_row.external_id
        assert task_data_row['row_data'] == data_row.row_data

        # data row details
        assert task_data_row_details['dataset_id'] == dataset.uid
        assert task_data_row_details['dataset_name'] == dataset.name

        assert task_data_row_details['last_activity_at'] is not None
        assert task_data_row_details['created_by'] is not None

        # media attributes
        assert task_media_attributes['mime_type'] == data_row.media_attributes[
            'mimeType']

        # project name and details
        assert task_project['name'] == project.name
        batch = next(project.batches())
        assert task_project_details['batch_id'] == batch.uid
        assert task_project_details['batch_name'] == batch.name
        assert task_project_details['priority'] is not None
        assert task_project_details[
            'consensus_expected_label_count'] is not None
        assert task_project_details['workflow_history'] is not None

        # label details
        assert task_project['labels'][0]['id'] == label_id


def test_project_export_v2_date_filters(client, export_v2_test_helpers,
                                        configured_project_with_label,
                                        wait_for_data_row_processing):
    project, _, data_row, label = configured_project_with_label
    data_row = wait_for_data_row_processing(client, data_row)
    label_id = label.uid

    task_name = "test_label_export_v2_date_filters"

    filters = {
        "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        "task_queue_status": "InReview"
    }

    # TODO: Right now we don't have a way to test this
    include_performance_details = True
    params = {
        "performance_details": include_performance_details,
        "include_labels": True,
        "project_details": True,
        "media_type_override": MediaType.Image
    }

    task_queues = project.task_queues()

    review_queue = next(
        tq for tq in task_queues if tq.queue_type == "MANUAL_REVIEW_QUEUE")
    project.move_data_rows_to_task_queue([data_row.uid], review_queue.uid)

    task_results = export_v2_test_helpers.run_project_export_v2_task(
        project, task_name=task_name, filters=filters, params=params)

    for task_result in task_results:
        task_project = task_result['projects'][project.uid]
        task_project_label_ids_set = set(
            map(lambda prediction: prediction['id'], task_project['labels']))
        assert label_id in task_project_label_ids_set
        assert task_project['project_details']['workflow_status'] == 'IN_REVIEW'

        # TODO: Add back in when we have a way to test this
        # if include_performance_details:
        #     assert 'include_performance_details' in task_result and task_result[
        #         'include_performance_details'] is not None
        # else:
        #     assert 'include_performance_details' not in task_result or task_result[
        #         'include_performance_details'] is None

    filters = {"last_activity_at": [None, "2050-01-01 00:00:00"]}
    export_v2_test_helpers.run_project_export_v2_task(project, filters=filters)

    filters = {"label_created_at": ["2000-01-01 00:00:00", None]}
    export_v2_test_helpers.run_project_export_v2_task(project, filters=filters)


def test_project_export_v2_with_iso_date_filters(client, export_v2_test_helpers,
                                                 configured_project_with_label,
                                                 wait_for_data_row_processing):
    project, _, data_row, label = configured_project_with_label
    data_row = wait_for_data_row_processing(client, data_row)
    label_id = label.uid

    task_name = "test_label_export_v2_with_iso_date_filters"

    filters = {
        "last_activity_at": [
            "2000-01-01T00:00:00+0230", "2050-01-01T00:00:00+0230"
        ],
        "label_created_at": [
            "2000-01-01T00:00:00+0230", "2050-01-01T00:00:00+0230"
        ]
    }
    task_results = export_v2_test_helpers.run_project_export_v2_task(
        project, task_name=task_name, filters=filters)
    assert label_id == task_results[0]['projects'][
        project.uid]['labels'][0]['id']

    filters = {"last_activity_at": [None, "2050-01-01T00:00:00+0230"]}
    task_results = export_v2_test_helpers.run_project_export_v2_task(
        project, task_name=task_name, filters=filters)
    assert label_id == task_results[0]['projects'][
        project.uid]['labels'][0]['id']

    filters = {"label_created_at": ["2050-01-01T00:00:00+0230", None]}
    task_results = export_v2_test_helpers.run_project_export_v2_task(
        project, task_name=task_name, filters=filters)
    assert len(task_results) == 0


@pytest.mark.parametrize("data_rows", [3], indirect=True)
def test_project_export_v2_datarows_filter(
        export_v2_test_helpers,
        configured_batch_project_with_multiple_datarows):
    project, _, data_rows = configured_batch_project_with_multiple_datarows

    data_row_ids = [dr.uid for dr in data_rows]
    datarow_filter_size = 2

    filters = {
        "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        "data_row_ids": data_row_ids[:datarow_filter_size]
    }
    params = {"data_row_details": True, "media_type_override": MediaType.Image}
    task_results = export_v2_test_helpers.run_project_export_v2_task(
        project, filters=filters, params=params)

    # only 2 datarows should be exported
    assert len(task_results) == datarow_filter_size
    # only filtered datarows should be exported
    assert set([dr['data_row']['id'] for dr in task_results
               ]) == set(data_row_ids[:datarow_filter_size])

    global_keys = [dr.global_key for dr in data_rows]
    filters = {
        "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        "global_keys": global_keys[:datarow_filter_size]
    }
    params = {"data_row_details": True, "media_type_override": MediaType.Image}
    task_results = export_v2_test_helpers.run_project_export_v2_task(
        project, filters=filters, params=params)

    # only 2 datarows should be exported
    assert len(task_results) == datarow_filter_size
    # only filtered datarows should be exported
    assert set([dr['data_row']['global_key'] for dr in task_results
               ]) == set(global_keys[:datarow_filter_size])


def test_batch_project_export_v2(
        configured_batch_project_with_label: Tuple[Project, Dataset, DataRow,
                                                   Label],
        export_v2_test_helpers, dataset: Dataset, image_url: str):
    project, dataset, *_ = configured_batch_project_with_label

    batch = list(project.batches())[0]
    filters = {
        "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        "batch_ids": [batch.uid],
    }
    params = {
        "include_performance_details": True,
        "include_labels": True,
        "media_type_override": MediaType.Image
    }
    task_name = "test_batch_export_v2"
    task = dataset.create_data_rows([
        {
            "row_data": image_url,
            "external_id": "my-image"
        },
    ] * 2)
    task.wait_till_done()
    data_rows = [dr.uid for dr in list(dataset.export_data_rows())]
    batch_one = f'batch one {uuid.uuid4()}'

    # This test creates two batches, only one batch should be exporter
    # Creatin second batch that will not be used in the export due to the filter: batch_id
    project.create_batch(batch_one, data_rows)

    task_results = export_v2_test_helpers.run_project_export_v2_task(
        project, task_name=task_name, filters=filters, params=params)
    assert (batch.size == len(task_results))



from unittest.mock import MagicMock

from labelbox.schema.export_task import (
    Converter,
    FileConverter,
    Range,
    StreamType,
    _MetadataFileInfo,
    _MetadataHeader,
    _TaskContext,
)


class TestFileConverter:

    def test_with_correct_ndjson(self, tmp_path, generate_random_ndjson):
        directory = tmp_path / "file-converter"
        directory.mkdir()
        line_count = 10
        ndjson = generate_random_ndjson(line_count)
        file_content = "\n".join(ndjson) + "\n"
        input_args = Converter.ConverterInputArgs(
            ctx=_TaskContext(
                client=MagicMock(),
                task_id="task-id",
                stream_type=StreamType.RESULT,
                metadata_header=_MetadataHeader(total_size=len(file_content),
                                                total_lines=line_count),
            ),
            file_info=_MetadataFileInfo(
                offsets=Range(start=0, end=len(file_content) - 1),
                lines=Range(start=0, end=line_count - 1),
                file="file.ndjson",
            ),
            raw_data=file_content,
        )
        path = directory / "output.ndjson"
        with FileConverter(file_path=path) as converter:
            for output in converter.convert(input_args):
                assert output.current_line == 0
                assert output.current_offset == 0
                assert output.file_path == path
                assert output.total_lines == line_count
                assert output.total_size == len(file_content)
                assert output.bytes_written == len(file_content)

    def test_with_no_newline_at_end(self, tmp_path, generate_random_ndjson):
        directory = tmp_path / "file-converter"
        directory.mkdir()
        line_count = 10
        ndjson = generate_random_ndjson(line_count)
        file_content = "\n".join(ndjson)
        input_args = Converter.ConverterInputArgs(
            ctx=_TaskContext(
                client=MagicMock(),
                task_id="task-id",
                stream_type=StreamType.RESULT,
                metadata_header=_MetadataHeader(total_size=len(file_content),
                                                total_lines=line_count),
            ),
            file_info=_MetadataFileInfo(
                offsets=Range(start=0, end=len(file_content) - 1),
                lines=Range(start=0, end=line_count - 1),
                file="file.ndjson",
            ),
            raw_data=file_content,
        )
        path = directory / "output.ndjson"
        with FileConverter(file_path=path) as converter:
            for output in converter.convert(input_args):
                assert output.current_line == 0
                assert output.current_offset == 0
                assert output.file_path == path
                assert output.total_lines == line_count
                assert output.total_size == len(file_content)
                assert output.bytes_written == len(file_content)



from typing import Optional
from uuid import uuid4

from labelbox.utils import _CamelCaseMixin, is_exactly_one_set
from labelbox import pydantic_compat
from ...annotation_types.types import Cuid


class DataRow(_CamelCaseMixin):
    id: str = None
    global_key: str = None

    @pydantic_compat.root_validator()
    def must_set_one(cls, values):
        if not is_exactly_one_set(values.get('id'), values.get('global_key')):
            raise ValueError("Must set either id or global_key")
        return values


class NDJsonBase(_CamelCaseMixin):
    uuid: str = None
    data_row: DataRow

    @pydantic_compat.validator('uuid', pre=True, always=True)
    def set_id(cls, v):
        return v or str(uuid4())

    def dict(self, *args, **kwargs):
        """ Pop missing id or missing globalKey from dataRow """
        res = super().dict(*args, **kwargs)
        if not self.data_row.id:
            res['dataRow'].pop('id')
        if not self.data_row.global_key:
            res['dataRow'].pop('globalKey')
        return res


class NDAnnotation(NDJsonBase):
    name: Optional[str] = None
    schema_id: Optional[Cuid] = None
    message_id: Optional[str] = None
    page: Optional[int] = None
    unit: Optional[str] = None

    @pydantic_compat.root_validator()
    def must_set_one(cls, values):
        if ('schema_id' not in values or values['schema_id']
                is None) and ('name' not in values or values['name'] is None):
            raise ValueError("Schema id or name are not set. Set either one.")
        return values

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        if 'name' in res and res['name'] is None:
            res.pop('name')
        if 'schemaId' in res and res['schemaId'] is None:
            res.pop('schemaId')
        return res



from labelbox.data.annotation_types.annotation import ClassificationAnnotation
from labelbox.data.annotation_types.classification.classification import Checklist, ClassificationAnswer, Radio, Text
from labelbox.data.annotation_types.data.text import TextData
from labelbox.data.annotation_types.label import Label

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_serialization():
    label = Label(uid="ckj7z2q0b0000jx6x0q2q7q0d",
                  data=TextData(
                      uid="bkj7z2q0b0000jx6x0q2q7q0d",
                      text="This is a test",
                  ),
                  annotations=[
                      ClassificationAnnotation(name="free_text_annotation",
                                               value=Text(confidence=0.5,
                                                          answer="text_answer"))
                  ])

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)

    assert res['confidence'] == 0.5
    assert res['name'] == "free_text_annotation"
    assert res['answer'] == "text_answer"
    assert res['dataRow']['id'] == "bkj7z2q0b0000jx6x0q2q7q0d"

    deserialized = NDJsonConverter.deserialize([res])
    res = next(deserialized)

    annotation = res.annotations[0]

    annotation_value = annotation.value
    assert type(annotation_value) is Text
    assert annotation_value.answer == "text_answer"
    assert annotation_value.confidence == 0.5


def test_nested_serialization():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=TextData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
            text="This is a test",
        ),
        annotations=[
            ClassificationAnnotation(
                name="nested test",
                value=Checklist(answer=[
                    ClassificationAnswer(
                        name="first_answer",
                        confidence=0.9,
                        classifications=[
                            ClassificationAnnotation(
                                name="sub_radio_question",
                                value=Radio(answer=ClassificationAnswer(
                                    name="first_sub_radio_answer",
                                    confidence=0.8,
                                    classifications=[
                                        ClassificationAnnotation(
                                            name="nested answer",
                                            value=Text(
                                                answer="nested answer",
                                                confidence=0.7,
                                            ))
                                    ])))
                        ])
                ]),
            )
        ])

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)

    assert res['dataRow']['id'] == "bkj7z2q0b0000jx6x0q2q7q0d"
    answer = res['answer'][0]
    assert answer['confidence'] == 0.9
    assert answer['name'] == "first_answer"
    classification = answer['classifications'][0]
    nested_classification_answer = classification['answer']
    assert nested_classification_answer['confidence'] == 0.8
    assert nested_classification_answer['name'] == "first_sub_radio_answer"
    sub_classification = nested_classification_answer['classifications'][0]
    assert sub_classification['name'] == "nested answer"
    assert sub_classification['answer'] == "nested answer"
    assert sub_classification['confidence'] == 0.7

    deserialized = NDJsonConverter.deserialize([res])
    res = next(deserialized)
    annotation = res.annotations[0]
    answer = annotation.value.answer[0]
    assert answer.confidence == 0.9
    assert answer.name == "first_answer"

    classification_answer = answer.classifications[0].value.answer
    assert classification_answer.confidence == 0.8
    assert classification_answer.name == "first_sub_radio_answer"

    sub_classification_answer = classification_answer.classifications[0].value
    assert type(sub_classification_answer) is Text
    assert sub_classification_answer.answer == "nested answer"
    assert sub_classification_answer.confidence == 0.7



from copy import copy
import pytest
import base64
import labelbox.types as lb_types
from labelbox.data.serialization import NDJsonConverter
from labelbox.data.serialization.ndjson.objects import NDDicomSegments, NDDicomSegment, NDDicomLine
"""
Polyline test data
"""

dicom_polyline_annotations = [
    lb_types.DICOMObjectAnnotation(uuid="78a8a027-9089-420c-8348-6099eb77e4aa",
                                   name="dicom_polyline",
                                   frame=2,
                                   value=lb_types.Line(points=[
                                       lb_types.Point(x=680, y=100),
                                       lb_types.Point(x=100, y=190),
                                       lb_types.Point(x=190, y=220)
                                   ]),
                                   segment_index=0,
                                   keyframe=True,
                                   group_key=lb_types.GroupKey.AXIAL)
]

polyline_label = lb_types.Label(data=lb_types.DicomData(uid="test-uid"),
                                annotations=dicom_polyline_annotations)

polyline_annotation_ndjson = {
    'classifications': [],
    'dataRow': {
        'id': 'test-uid'
    },
    'name':
        'dicom_polyline',
    'groupKey':
        'axial',
    'segments': [{
        'keyframes': [{
            'frame': 2,
            'line': [
                {
                    'x': 680.0,
                    'y': 100.0
                },
                {
                    'x': 100.0,
                    'y': 190.0
                },
                {
                    'x': 190.0,
                    'y': 220.0
                },
            ],
            'classifications': [],
        }]
    }],
}

polyline_with_global_key = lb_types.Label(
    data=lb_types.DicomData(global_key="test-global-key"),
    annotations=dicom_polyline_annotations)

polyline_annotation_ndjson_with_global_key = copy(polyline_annotation_ndjson)
polyline_annotation_ndjson_with_global_key['dataRow'] = {
    'globalKey': 'test-global-key'
}
"""
Video test data
"""

instance_uri_1 = 'https://storage.labelbox.com/cjhfn5y6s0pk507024nz1ocys%2F1d60856c-59b7-3060-2754-83f7e93e0d01-1?Expires=1666901963361&KeyName=labelbox-assets-key-3&Signature=t-2s2DB4YjFuWEFak0wxYqfBfZA'
instance_uri_5 = 'https://storage.labelbox.com/cjhfn5y6s0pk507024nz1ocys1%2F1d60856c-59b7-3060-2754-83f7e93e0d01-1?Expires=1666901963361&KeyName=labelbox-assets-key-3&Signature=t-2s2DB4YjFuWEFak0wxYqfBfZA'
frames = [
    lb_types.MaskFrame(index=1, instance_uri=instance_uri_1),
    lb_types.MaskFrame(index=5, instance_uri=instance_uri_5)
]
instances = [
    lb_types.MaskInstance(color_rgb=(0, 0, 255), name="mask1"),
    lb_types.MaskInstance(color_rgb=(0, 255, 0), name="mask2"),
    lb_types.MaskInstance(color_rgb=(255, 0, 0), name="mask3")
]

video_mask_annotation = lb_types.VideoMaskAnnotation(frames=frames,
                                                     instances=instances)

video_mask_annotation_ndjson = {
    'dataRow': {
        'id': 'test-uid'
    },
    'masks': {
        'frames': [{
            'index': 1,
            'imBytes': None,
            'instanceURI': instance_uri_1
        }, {
            'index': 5,
            'imBytes': None,
            'instanceURI': instance_uri_5
        }],
        'instances': [
            {
                'colorRGB': (0, 0, 255),
                'name': 'mask1'
            },
            {
                'colorRGB': (0, 255, 0),
                'name': 'mask2'
            },
            {
                'colorRGB': (255, 0, 0),
                'name': 'mask3'
            },
        ]
    },
}

video_mask_annotation_ndjson_with_global_key = copy(
    video_mask_annotation_ndjson)
video_mask_annotation_ndjson_with_global_key['dataRow'] = {
    'globalKey': 'test-global-key'
}

video_mask_label = lb_types.Label(data=lb_types.VideoData(uid="test-uid"),
                                  annotations=[video_mask_annotation])

video_mask_label_with_global_key = lb_types.Label(
    data=lb_types.VideoData(global_key="test-global-key"),
    annotations=[video_mask_annotation])
"""
DICOM Mask test data
"""

dicom_mask_annotation = lb_types.DICOMMaskAnnotation(
    name="dicom_mask",
    group_key=lb_types.GroupKey.AXIAL,
    frames=frames,
    instances=instances)

dicom_mask_label = lb_types.Label(data=lb_types.DicomData(uid="test-uid"),
                                  annotations=[dicom_mask_annotation])

dicom_mask_label_with_global_key = lb_types.Label(
    data=lb_types.DicomData(global_key="test-global-key"),
    annotations=[dicom_mask_annotation])

dicom_mask_annotation_ndjson = copy(video_mask_annotation_ndjson)
dicom_mask_annotation_ndjson['groupKey'] = 'axial'
dicom_mask_annotation_ndjson_with_global_key = copy(
    dicom_mask_annotation_ndjson)
dicom_mask_annotation_ndjson_with_global_key['dataRow'] = {
    'globalKey': 'test-global-key'
}
"""
Tests
"""

labels = [
    polyline_label, polyline_with_global_key, dicom_mask_label,
    dicom_mask_label_with_global_key, video_mask_label,
    video_mask_label_with_global_key
]
ndjsons = [
    polyline_annotation_ndjson,
    polyline_annotation_ndjson_with_global_key,
    dicom_mask_annotation_ndjson,
    dicom_mask_annotation_ndjson_with_global_key,
    video_mask_annotation_ndjson,
    video_mask_annotation_ndjson_with_global_key,
]
labels_ndjsons = list(zip(labels, ndjsons))


def test_deserialize_nd_dicom_segments():
    nd_dicom_segments = NDDicomSegments(**polyline_annotation_ndjson)
    assert isinstance(nd_dicom_segments, NDDicomSegments)
    assert isinstance(nd_dicom_segments.segments[0], NDDicomSegment)
    assert isinstance(nd_dicom_segments.segments[0].keyframes[0], NDDicomLine)


@pytest.mark.parametrize('label, ndjson', labels_ndjsons)
def test_serialize_label(label, ndjson):
    serialized_label = next(NDJsonConverter().serialize([label]))
    serialized_label.pop('uuid')
    assert serialized_label == ndjson


@pytest.mark.parametrize('label, ndjson', labels_ndjsons)
def test_deserialize_label(label, ndjson):
    deserialized_label = next(NDJsonConverter().deserialize([ndjson]))
    if hasattr(deserialized_label.annotations[0], 'extra'):
        deserialized_label.annotations[0].extra = {}
    assert deserialized_label.annotations == label.annotations


@pytest.mark.parametrize('label', labels)
def test_serialize_deserialize_label(label):
    serialized = list(NDJsonConverter.serialize([label]))
    deserialized = list(NDJsonConverter.deserialize(serialized))
    if hasattr(deserialized[0].annotations[0], 'extra'):
        deserialized[0].annotations[0].extra = {}
    assert deserialized[0].annotations == label.annotations



from labelbox.data.annotation_types.annotation import ClassificationAnnotation
from labelbox.data.annotation_types.classification.classification import Checklist, ClassificationAnswer, Radio
from labelbox.data.annotation_types.data.text import TextData
from labelbox.data.annotation_types.label import Label

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_serialization_min():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=TextData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
            text="This is a test",
        ),
        annotations=[
            ClassificationAnnotation(
                name="checkbox_question_geo",
                value=Checklist(
                    answer=[ClassificationAnswer(name="first_answer")]),
            )
        ])

    expected = {
        'name': 'checkbox_question_geo',
        'dataRow': {
            'id': 'bkj7z2q0b0000jx6x0q2q7q0d'
        },
        'answer': [{
            'name': 'first_answer'
        }]
    }
    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)
    res.pop("uuid")
    assert res == expected

    deserialized = NDJsonConverter.deserialize([res])
    res = next(deserialized)
    res.annotations[0].extra.pop("uuid")
    assert res.annotations == label.annotations


def test_serialization_with_classification():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=TextData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
            text="This is a test",
        ),
        annotations=[
            ClassificationAnnotation(
                name="checkbox_question_geo",
                confidence=0.5,
                value=Checklist(answer=[
                    ClassificationAnswer(
                        name="first_answer",
                        confidence=0.1,
                        classifications=[
                            ClassificationAnnotation(
                                name="sub_radio_question",
                                value=Radio(answer=ClassificationAnswer(
                                    name="first_sub_radio_answer",
                                    confidence=0.31))),
                            ClassificationAnnotation(
                                name="sub_chck_question",
                                value=Checklist(answer=[
                                    ClassificationAnswer(
                                        name="second_subchk_answer",
                                        confidence=0.41),
                                    ClassificationAnswer(
                                        name="third_subchk_answer",
                                        confidence=0.42),
                                ],))
                        ]),
                ]))
        ])

    expected = {
        'confidence':
            0.5,
        'name':
            'checkbox_question_geo',
        'dataRow': {
            'id': 'bkj7z2q0b0000jx6x0q2q7q0d'
        },
        'answer': [{
            'confidence':
                0.1,
            'name':
                'first_answer',
            'classifications': [{
                'name': 'sub_radio_question',
                'answer': {
                    'confidence': 0.31,
                    'name': 'first_sub_radio_answer',
                }
            }, {
                'name':
                    'sub_chck_question',
                'answer': [{
                    'confidence': 0.41,
                    'name': 'second_subchk_answer',
                }, {
                    'confidence': 0.42,
                    'name': 'third_subchk_answer',
                }]
            }]
        }]
    }

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)

    res.pop("uuid")
    assert res == expected

    deserialized = NDJsonConverter.deserialize([res])
    res = next(deserialized)
    res.annotations[0].extra.pop("uuid")
    assert res.annotations == label.annotations


def test_serialization_with_classification_double_nested():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=TextData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
            text="This is a test",
        ),
        annotations=[
            ClassificationAnnotation(
                name="checkbox_question_geo",
                confidence=0.5,
                value=Checklist(answer=[
                    ClassificationAnswer(
                        name="first_answer",
                        confidence=0.1,
                        classifications=[
                            ClassificationAnnotation(
                                name="sub_radio_question",
                                value=Radio(answer=ClassificationAnswer(
                                    name="first_sub_radio_answer",
                                    confidence=0.31,
                                    classifications=[
                                        ClassificationAnnotation(
                                            name="sub_chck_question",
                                            value=Checklist(answer=[
                                                ClassificationAnswer(
                                                    name="second_subchk_answer",
                                                    confidence=0.41),
                                                ClassificationAnswer(
                                                    name="third_subchk_answer",
                                                    confidence=0.42),
                                            ],))
                                    ]))),
                        ]),
                ]))
        ])

    expected = {
        'confidence':
            0.5,
        'name':
            'checkbox_question_geo',
        'dataRow': {
            'id': 'bkj7z2q0b0000jx6x0q2q7q0d'
        },
        'answer': [{
            'confidence':
                0.1,
            'name':
                'first_answer',
            'classifications': [{
                'name': 'sub_radio_question',
                'answer': {
                    'confidence':
                        0.31,
                    'name':
                        'first_sub_radio_answer',
                    'classifications': [{
                        'name':
                            'sub_chck_question',
                        'answer': [{
                            'confidence': 0.41,
                            'name': 'second_subchk_answer',
                        }, {
                            'confidence': 0.42,
                            'name': 'third_subchk_answer',
                        }]
                    }]
                }
            }]
        }]
    }
    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)

    res.pop("uuid")
    assert res == expected

    deserialized = NDJsonConverter.deserialize([res])
    res = next(deserialized)
    res.annotations[0].extra.pop("uuid")
    assert res.annotations == label.annotations


def test_serialization_with_classification_double_nested_2():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=TextData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
            text="This is a test",
        ),
        annotations=[
            ClassificationAnnotation(
                name="sub_radio_question",
                value=Radio(answer=ClassificationAnswer(
                    name="first_sub_radio_answer",
                    confidence=0.31,
                    classifications=[
                        ClassificationAnnotation(
                            name="sub_chck_question",
                            value=Checklist(answer=[
                                ClassificationAnswer(
                                    name="second_subchk_answer",
                                    confidence=0.41,
                                    classifications=[
                                        ClassificationAnnotation(
                                            name="checkbox_question_geo",
                                            value=Checklist(answer=[
                                                ClassificationAnswer(
                                                    name="first_answer",
                                                    confidence=0.1,
                                                    classifications=[]),
                                            ]))
                                    ]),
                                ClassificationAnswer(name="third_subchk_answer",
                                                     confidence=0.42),
                            ]))
                    ]))),
        ])

    expected = {
        'name': 'sub_radio_question',
        'answer': {
            'confidence':
                0.31,
            'name':
                'first_sub_radio_answer',
            'classifications': [{
                'name':
                    'sub_chck_question',
                'answer': [{
                    'confidence':
                        0.41,
                    'name':
                        'second_subchk_answer',
                    'classifications': [{
                        'name': 'checkbox_question_geo',
                        'answer': [{
                            'confidence': 0.1,
                            'name': 'first_answer',
                        }]
                    }]
                }, {
                    'confidence': 0.42,
                    'name': 'third_subchk_answer',
                }]
            }]
        },
        'dataRow': {
            'id': 'bkj7z2q0b0000jx6x0q2q7q0d'
        }
    }

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)
    res.pop("uuid")
    assert res == expected

    deserialized = NDJsonConverter.deserialize([res])
    res = next(deserialized)
    res.annotations[0].extra.pop("uuid")
    assert res.annotations == label.annotations



import json
import random

from labelbox import StreamType, JsonConverter


class TestExportEmbeddings:

    def test_export_embeddings_precomputed(self, client, dataset, environ,
                                           image_url):
        data_row_specs = [{
            "row_data": image_url,
            "external_id": "image",
        }]
        task = dataset.create_data_rows(data_row_specs)
        task.wait_till_done()
        export_task = dataset.export(params={"embeddings": True})
        export_task.wait_till_done()
        assert export_task.status == "COMPLETE"
        assert export_task.has_result()
        assert export_task.has_errors() is False

        results = []
        export_task.get_stream(converter=JsonConverter(),
                               stream_type=StreamType.RESULT).start(
                                   stream_handler=lambda output: results.append(
                                       json.loads(output.json_str)))

        assert len(results) == len(data_row_specs)

        result = results[0]
        assert "embeddings" in result
        assert len(result["embeddings"]) > 0
        assert result["embeddings"][0][
            "name"] == "Image Embedding V2 (CLIP ViT-B/32)"
        assert len(result["embeddings"][0]["values"]) == 1

    def test_export_embeddings_custom(self, client, dataset, image_url,
                                      embedding):
        vector = [random.uniform(1.0, 2.0) for _ in range(embedding.dims)]
        import_task = dataset.create_data_rows([{
            "row_data": image_url,
            "embeddings": [{
                "embedding_id": embedding.id,
                "vector": vector,
            }],
        }])
        import_task.wait_till_done()
        assert import_task.status == "COMPLETE"

        export_task = dataset.export(params={"embeddings": True})
        export_task.wait_till_done()
        assert export_task.status == "COMPLETE"
        assert export_task.has_result()
        assert export_task.has_errors() is False

        results = []
        export_task.get_stream(converter=JsonConverter(),
                               stream_type=StreamType.RESULT).start(
                                   stream_handler=lambda output: results.append(
                                       json.loads(output.json_str)))

        assert len(results) == 1
        assert "embeddings" in results[0]
        assert (len(results[0]["embeddings"])
                >= 1)  # should at least contain the custom embedding
        for emb in results[0]["embeddings"]:
            if emb["id"] == embedding.id:
                assert emb["name"] == embedding.name
                assert emb["dimensions"] == embedding.dims
                assert emb["is_custom"] == True
                assert len(emb["values"]) == 1
                assert emb["values"][0]["value"] == vector



from labelbox.typing_imports import Literal
from labelbox.utils import _NoCoercionMixin
from .base_data import BaseData


class DicomData(BaseData, _NoCoercionMixin):
    class_name: Literal["DicomData"] = "DicomData"


from typing import Any, Dict, List, Union, Optional
import warnings
from labelbox.data.annotation_types.base_annotation import BaseAnnotation

from labelbox.data.mixins import ConfidenceMixin, CustomMetricsMixin

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from labelbox import pydantic_compat
from ..feature import FeatureSchema


# TODO: Replace when pydantic adds support for unions that don't coerce types
class _TempName(ConfidenceMixin, pydantic_compat.BaseModel):
    name: str

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        res.pop('name')
        return res


class ClassificationAnswer(FeatureSchema, ConfidenceMixin, CustomMetricsMixin):
    """
    - Represents a classification option.
    - Because it inherits from FeatureSchema
        the option can be represented with either the name or feature_schema_id

    - The keyframe arg only applies to video classifications.
      Each answer can have a keyframe independent of the others.
        So unlike object annotations, classification annotations
          track keyframes at a classification answer level.
    """
    extra: Dict[str, Any] = {}
    keyframe: Optional[bool] = None
    classifications: List['ClassificationAnnotation'] = []

    def dict(self, *args, **kwargs) -> Dict[str, str]:
        res = super().dict(*args, **kwargs)
        if res['keyframe'] is None:
            res.pop('keyframe')
        if res['classifications'] == []:
            res.pop('classifications')
        return res


class Radio(ConfidenceMixin, CustomMetricsMixin, pydantic_compat.BaseModel):
    """ A classification with only one selected option allowed

    >>> Radio(answer = ClassificationAnswer(name = "dog"))

    """
    answer: ClassificationAnswer


class Checklist(_TempName):
    """ A classification with many selected options allowed

    >>> Checklist(answer = [ClassificationAnswer(name = "cloudy")])

    """
    name: Literal["checklist"] = "checklist"
    answer: List[ClassificationAnswer]


class Text(ConfidenceMixin, CustomMetricsMixin, pydantic_compat.BaseModel):
    """ Free form text

    >>> Text(answer = "some text answer")

    """
    answer: str


class Dropdown(_TempName):
    """
    - A classification with many selected options allowed .
    - This is not currently compatible with MAL.

    Deprecation Notice: Dropdown classification is deprecated and will be
        removed in a future release. Dropdown will also
        no longer be able to be created in the Editor on 3/31/2022.    
    """
    name: Literal["dropdown"] = "dropdown"
    answer: List[ClassificationAnswer]

    def __init__(self, **data: Any):
        super().__init__(**data)
        warnings.warn("Dropdown classification is deprecated and will be "
                      "removed in a future release")


class ClassificationAnnotation(BaseAnnotation, ConfidenceMixin,
                               CustomMetricsMixin):
    """Classification annotations (non localized)

    >>> ClassificationAnnotation(
    >>>     value=Text(answer="my caption message"),
    >>>     feature_schema_id="my-feature-schema-id"
    >>> )

    Args:
        name (Optional[str])
        classifications (Optional[List[ClassificationAnnotation]]): Optional sub classification of the annotation
        feature_schema_id (Optional[Cuid])
        value (Union[Text, Checklist, Radio, Dropdown])
        extra (Dict[str, Any])
     """

    value: Union[Text, Checklist, Radio, Dropdown]
    message_id: Optional[str] = None


ClassificationAnswer.update_forward_refs()



from enum import Enum
from typing import List, Optional, Tuple

from labelbox import pydantic_compat
from labelbox.data.annotation_types.annotation import ClassificationAnnotation, ObjectAnnotation

from labelbox.data.annotation_types.annotation import ClassificationAnnotation, ObjectAnnotation
from labelbox.data.annotation_types.feature import FeatureSchema
from labelbox.data.mixins import ConfidenceNotSupportedMixin, CustomMetricsNotSupportedMixin
from labelbox.utils import _CamelCaseMixin, is_valid_uri


class VideoClassificationAnnotation(ClassificationAnnotation):
    """Video classification
    Args:
        name (Optional[str])
        feature_schema_id (Optional[Cuid])
        value (Union[Text, Checklist, Radio, Dropdown])
        frame (int): The frame index that this annotation corresponds to
        segment_id (Optional[Int]): Index of video segment this annotation belongs to
        extra (Dict[str, Any])
    """
    frame: int
    segment_index: Optional[int] = None


class VideoObjectAnnotation(ObjectAnnotation, ConfidenceNotSupportedMixin,
                            CustomMetricsNotSupportedMixin):
    """Video object annotation
    >>> VideoObjectAnnotation(
    >>>     keyframe=True,
    >>>     frame=10,
    >>>     value=Rectangle(
    >>>        start=Point(x=0, y=0),
    >>>        end=Point(x=1, y=1)
    >>>     ),
    >>>     feature_schema_id="my-feature-schema-id"
    >>> )
    Args:
        name (Optional[str])
        feature_schema_id (Optional[Cuid])
        value (Geometry)
        frame (Int): The frame index that this annotation corresponds to
        keyframe (bool): Whether or not this annotation was a human generated or interpolated annotation
        segment_id (Optional[Int]): Index of video segment this annotation belongs to
        classifications (List[ClassificationAnnotation]) = []
        extra (Dict[str, Any])
    """
    frame: int
    keyframe: bool
    segment_index: Optional[int] = None


class GroupKey(Enum):
    """Group key for DICOM annotations
    """
    AXIAL = "axial"
    SAGITTAL = "sagittal"
    CORONAL = "coronal"


class DICOMObjectAnnotation(VideoObjectAnnotation):
    """DICOM object annotation
    >>> DICOMObjectAnnotation(
    >>>     name="dicom_polyline",
    >>>     frame=2,
    >>>     value=lb_types.Line(points = [
    >>>         lb_types.Point(x=680, y=100),
    >>>         lb_types.Point(x=100, y=190),
    >>>         lb_types.Point(x=190, y=220)
    >>>     ]),
    >>>     segment_index=0,
    >>>     keyframe=True,
    >>>     Group_key=GroupKey.AXIAL
    >>> )
    Args:
        name (Optional[str])
        feature_schema_id (Optional[Cuid])
        value (Geometry)
        group_key (GroupKey)
        frame (Int): The frame index that this annotation corresponds to
        keyframe (bool): Whether or not this annotation was a human generated or interpolated annotation
        segment_id (Optional[Int]): Index of video segment this annotation belongs to
        classifications (List[ClassificationAnnotation]) = []
        extra (Dict[str, Any])
    """
    group_key: GroupKey


class MaskFrame(_CamelCaseMixin, pydantic_compat.BaseModel):
    index: int
    instance_uri: Optional[str] = None
    im_bytes: Optional[bytes] = None

    @pydantic_compat.root_validator()
    def validate_args(cls, values):
        im_bytes = values.get("im_bytes")
        instance_uri = values.get("instance_uri")

        if im_bytes == instance_uri == None:
            raise ValueError("One of `instance_uri`, `im_bytes` required.")
        return values

    @pydantic_compat.validator("instance_uri")
    def validate_uri(cls, v):
        if not is_valid_uri(v):
            raise ValueError(f"{v} is not a valid uri")
        return v


class MaskInstance(_CamelCaseMixin, FeatureSchema):
    color_rgb: Tuple[int, int, int]
    name: str


class VideoMaskAnnotation(pydantic_compat.BaseModel):
    """Video mask annotation
       >>> VideoMaskAnnotation(
       >>>     frames=[
       >>>         MaskFrame(index=1, instance_uri='https://storage.labelbox.com/cjhfn5y6s0pk507024nz1ocys%2F1d60856c-59b7-3060-2754-83f7e93e0d01-1?Expires=1666901963361&KeyName=labelbox-assets-key-3&Signature=t-2s2DB4YjFuWEFak0wxYqfBfZA'),
       >>>         MaskFrame(index=5, instance_uri='https://storage.labelbox.com/cjhfn5y6s0pk507024nz1ocys1%2F1d60856c-59b7-3060-2754-83f7e93e0d01-1?Expires=1666901963361&KeyName=labelbox-assets-key-3&Signature=t-2s2DB4YjFuWEFak0wxYqfBfZA'),
       >>>     ],
       >>>     instances=[
       >>>         MaskInstance(color_rgb=(0, 0, 255), name="mask1"),
       >>>         MaskInstance(color_rgb=(0, 255, 0), name="mask2"),
       >>>         MaskInstance(color_rgb=(255, 0, 0), name="mask3")
       >>>     ]
       >>> )
    """
    frames: List[MaskFrame]
    instances: List[MaskInstance]


class DICOMMaskAnnotation(VideoMaskAnnotation):
    """DICOM mask annotation
       >>> DICOMMaskAnnotation(
       >>>     name="dicom_mask",
       >>>     group_key=GroupKey.AXIAL,
       >>>     frames=[
       >>>         MaskFrame(index=1, instance_uri='https://storage.labelbox.com/cjhfn5y6s0pk507024nz1ocys%2F1d60856c-59b7-3060-2754-83f7e93e0d01-1?Expires=1666901963361&KeyName=labelbox-assets-key-3&Signature=t-2s2DB4YjFuWEFak0wxYqfBfZA'),
       >>>         MaskFrame(index=5, instance_uri='https://storage.labelbox.com/cjhfn5y6s0pk507024nz1ocys1%2F1d60856c-59b7-3060-2754-83f7e93e0d01-1?Expires=1666901963361&KeyName=labelbox-assets-key-3&Signature=t-2s2DB4YjFuWEFak0wxYqfBfZA'),
       >>>     ],
       >>>     instances=[
       >>>         MaskInstance(color_rgb=(0, 0, 255), name="mask1"),
       >>>         MaskInstance(color_rgb=(0, 255, 0), name="mask2"),
       >>>         MaskInstance(color_rgb=(255, 0, 0), name="mask3")
       >>>     ]
       >>> )
    """
    group_key: GroupKey



name = "labelbox"

__version__ = "3.75.1"

from labelbox.client import Client
from labelbox.schema.project import Project
from labelbox.schema.model import Model
from labelbox.schema.model_config import ModelConfig
from labelbox.schema.bulk_import_request import BulkImportRequest
from labelbox.schema.annotation_import import MALPredictionImport, MEAPredictionImport, LabelImport, MEAToMALPredictionImport
from labelbox.schema.dataset import Dataset
from labelbox.schema.data_row import DataRow
from labelbox.schema.catalog import Catalog
from labelbox.schema.enums import AnnotationImportState
from labelbox.schema.label import Label
from labelbox.schema.batch import Batch
from labelbox.schema.review import Review
from labelbox.schema.user import User
from labelbox.schema.organization import Organization
from labelbox.schema.task import Task
from labelbox.schema.export_task import StreamType, ExportTask, JsonConverter, JsonConverterOutput, FileConverter, FileConverterOutput, BufferedJsonConverterOutput
from labelbox.schema.labeling_frontend import LabelingFrontend, LabelingFrontendOptions
from labelbox.schema.asset_attachment import AssetAttachment
from labelbox.schema.webhook import Webhook
from labelbox.schema.ontology import Ontology, OntologyBuilder, Classification, Option, Tool, FeatureSchema
from labelbox.schema.role import Role, ProjectRole
from labelbox.schema.invite import Invite, InviteLimit
from labelbox.schema.data_row_metadata import DataRowMetadataOntology, DataRowMetadataField, DataRowMetadata, DeleteDataRowMetadata
from labelbox.schema.model_run import ModelRun, DataSplit
from labelbox.schema.benchmark import Benchmark
from labelbox.schema.iam_integration import IAMIntegration
from labelbox.schema.resource_tag import ResourceTag
from labelbox.schema.project_model_config import ProjectModelConfig
from labelbox.schema.project_resource_tag import ProjectResourceTag
from labelbox.schema.media_type import MediaType
from labelbox.schema.slice import Slice, CatalogSlice, ModelSlice
from labelbox.schema.queue_mode import QueueMode
from labelbox.schema.task_queue import TaskQueue
from labelbox.schema.label_score import LabelScore
from labelbox.schema.identifiables import UniqueIds, GlobalKeys, DataRowIds
from labelbox.schema.identifiable import UniqueId, GlobalKey
from labelbox.schema.ontology_kind import OntologyKind
from labelbox.schema.project_overview import ProjectOverview, ProjectOverviewDetailed



import json
import uuid
from unittest.mock import patch

import pytest

from labelbox.schema.asset_attachment import AttachmentType
from labelbox.schema.identifiable import UniqueId, GlobalKey


class TestDataRowUpsert:

    @pytest.fixture
    def all_inclusive_data_row(self, dataset, image_url):
        dr = dataset.create_data_row(
            row_data=image_url,
            external_id="ex1",
            global_key=str(uuid.uuid4()),
            metadata_fields=[{
                "name": "tag",
                "value": "tag_string"
            }, {
                "name": "split",
                "value": "train"
            }],
            attachments=[
                {
                    "type": "RAW_TEXT",
                    "name": "att1",
                    "value": "test1"
                },
                {
                    "type":
                        "IMAGE",
                    "name":
                        "att2",
                    "value":
                        "https://storage.googleapis.com/labelbox-sample-datasets/Docs/disease_attachment.jpeg"
                },
                {
                    "type":
                        "PDF_URL",
                    "name":
                        "att3",
                    "value":
                        "https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483.pdf"
                },
            ])
        return dr

    @pytest.mark.order(1)
    def test_create_data_row_with_auto_key(self, dataset, image_url):
        task = dataset.upsert_data_rows([{'row_data': image_url}])
        task.wait_till_done()
        assert len(list(dataset.data_rows())) == 1

    def test_create_data_row_with_upsert(self, client, dataset, image_url):
        gkey = str(uuid.uuid4())
        task = dataset.upsert_data_rows([{
            'row_data':
                image_url,
            'global_key':
                gkey,
            'external_id':
                "ex1",
            'attachments': [{
                'type': AttachmentType.RAW_TEXT,
                'name': "att1",
                'value': "test1"
            }, {
                'type':
                    AttachmentType.IMAGE,
                'name':
                    "att2",
                'value':
                    "https://storage.googleapis.com/labelbox-sample-datasets/Docs/disease_attachment.jpeg"
            }, {
                'type':
                    AttachmentType.PDF_URL,
                'name':
                    "att3",
                'value':
                    "https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483.pdf"
            }],
            'metadata': [{
                'name': "tag",
                'value': "updated tag"
            }, {
                'name': "split",
                'value': "train"
            }]
        }])
        task.wait_till_done()
        assert task.status == "COMPLETE"
        dr = client.get_data_row_by_global_key(gkey)

        assert dr is not None
        assert dr.row_data == image_url
        assert dr.global_key == gkey
        assert dr.external_id == "ex1"

        attachments = list(dr.attachments())
        assert len(attachments) == 3
        assert attachments[0].attachment_name == "att1"
        assert attachments[0].attachment_type == AttachmentType.RAW_TEXT
        assert attachments[0].attachment_value == "test1"

        assert attachments[1].attachment_name == "att2"
        assert attachments[1].attachment_type == AttachmentType.IMAGE
        assert attachments[
            1].attachment_value == "https://storage.googleapis.com/labelbox-sample-datasets/Docs/disease_attachment.jpeg"

        assert attachments[2].attachment_name == "att3"
        assert attachments[2].attachment_type == AttachmentType.PDF_URL
        assert attachments[
            2].attachment_value == "https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483.pdf"

        assert len(dr.metadata_fields) == 2
        assert dr.metadata_fields[0]['name'] == "tag"
        assert dr.metadata_fields[0]['value'] == "updated tag"
        assert dr.metadata_fields[1]['name'] == "split"
        assert dr.metadata_fields[1]['value'] == "train"

    def test_update_data_row_fields_with_upsert(self, client, dataset,
                                                image_url):
        gkey = str(uuid.uuid4())
        dr = dataset.create_data_row(row_data=image_url,
                                     external_id="ex1",
                                     global_key=gkey)
        task = dataset.upsert_data_rows([{
            'key': UniqueId(dr.uid),
            'external_id': "ex1_updated",
            'global_key': f"{gkey}_updated"
        }])
        task.wait_till_done()
        assert task.status == "COMPLETE"
        dr = client.get_data_row(dr.uid)
        assert dr is not None
        assert dr.external_id == "ex1_updated"
        assert dr.global_key == f"{gkey}_updated"

    def test_update_data_row_fields_with_upsert_by_global_key(
            self, client, dataset, image_url):
        gkey = str(uuid.uuid4())
        dr = dataset.create_data_row(row_data=image_url,
                                     external_id="ex1",
                                     global_key=gkey)
        task = dataset.upsert_data_rows([{
            'key': GlobalKey(dr.global_key),
            'external_id': "ex1_updated",
            'global_key': f"{gkey}_updated"
        }])
        task.wait_till_done()
        assert task.status == "COMPLETE"
        dr = client.get_data_row(dr.uid)
        assert dr is not None
        assert dr.external_id == "ex1_updated"
        assert dr.global_key == f"{gkey}_updated"

    def test_update_attachments_with_upsert(self, client,
                                            all_inclusive_data_row, dataset):
        dr = all_inclusive_data_row
        task = dataset.upsert_data_rows([{
            'key':
                UniqueId(dr.uid),
            'row_data':
                dr.row_data,
            'attachments': [{
                'type': AttachmentType.RAW_TEXT,
                'name': "att1",
                'value': "test"
            }]
        }])
        task.wait_till_done()
        assert task.status == "COMPLETE"
        dr = client.get_data_row(dr.uid)
        assert dr is not None
        attachments = list(dr.attachments())
        assert len(attachments) == 1
        assert attachments[0].attachment_name == "att1"

    def test_update_metadata_with_upsert(self, client, all_inclusive_data_row,
                                         dataset):
        dr = all_inclusive_data_row
        task = dataset.upsert_data_rows([{
            'key':
                GlobalKey(dr.global_key),
            'row_data':
                dr.row_data,
            'metadata': [{
                'name': "tag",
                'value': "updated tag"
            }, {
                'name': "split",
                'value': "train"
            }]
        }])
        task.wait_till_done()
        assert task.status == "COMPLETE"
        dr = client.get_data_row(dr.uid)
        assert dr is not None
        assert len(dr.metadata_fields) == 2
        assert dr.metadata_fields[0]['name'] == "tag"
        assert dr.metadata_fields[0]['value'] == "updated tag"
        assert dr.metadata_fields[1]['name'] == "split"
        assert dr.metadata_fields[1]['value'] == "train"

    def test_multiple_chunks(self, client, dataset, image_url):
        mocked_chunk_size = 300
        with patch('labelbox.client.Client.upload_data',
                   wraps=client.upload_data) as spy_some_function:
            with patch('labelbox.schema.dataset.UPSERT_CHUNK_SIZE_BYTES',
                       new=mocked_chunk_size):
                task = dataset.upsert_data_rows([{
                    'row_data': image_url
                } for i in range(10)])
                task.wait_till_done()
                assert len(list(dataset.data_rows())) == 10
                assert spy_some_function.call_count == 11  # one per each data row + manifest

                first_call_args, _ = spy_some_function.call_args_list[0]
                first_chunk_content = first_call_args[0]
                data = json.loads(first_chunk_content)
                """
                Each chunk but the last will have 3 items, the last will have 1.
                As the chunks are processed async, we cannot be sure of the order of the calls.
                """
                assert len(data) in {1, 3}

                last_call_args, _ = spy_some_function.call_args_list[-1]
                manifest_content = last_call_args[0].decode('utf-8')
                data = json.loads(manifest_content)
                assert data['source'] == "SDK"
                assert data['item_count'] == 10
                assert len(data['chunk_uris']) == 10

    def test_upsert_embedded_row_data(self, dataset):
        pdf_url = "https://lb-test-data.s3.us-west-1.amazonaws.com/document-samples/0801.3483.pdf"
        task = dataset.upsert_data_rows([{
            'row_data': {
                "pdf_url":
                    pdf_url,
                "text_layer_url":
                    "https://lb-test-data.s3.us-west-1.amazonaws.com/document-samples/0801.3483-lb-textlayer.json"
            },
            'media_type': "PDF"
        }])
        task.wait_till_done()
        data_rows = list(dataset.data_rows())
        assert len(data_rows) == 1
        assert data_rows[0].row_data == pdf_url

    def test_upsert_duplicate_global_key_error(self, dataset, image_url):
        gkey = str(uuid.uuid4())
        task = dataset.upsert_data_rows([
            {
                'row_data': image_url,
                'global_key': gkey
            },
            {
                'row_data': image_url,
                'global_key': gkey
            },
        ])
        task.wait_till_done()
        assert task.status == "COMPLETE"
        assert task.errors is not None
        assert len(task.errors) == 1  # one data row was created, one failed
        assert f"Duplicate global key: '{gkey}'" in task.errors[0]['message']

    def test_upsert_empty_items(self, dataset):
        items = [{"key": GlobalKey("foo")}]
        with pytest.raises(ValueError) as e_info:
            dataset.upsert_data_rows(items)
        e_info.match(r"The following items have an empty payload: \['foo'\]")



import json
import labelbox.types as lb_types
from labelbox.data.serialization.ndjson.converter import NDJsonConverter

bbox_annotation = lb_types.ObjectAnnotation(
    name="bounding_box",  # must match your ontology feature's name
    value=lb_types.DocumentRectangle(
        start=lb_types.Point(x=42.799, y=86.498),  # Top left
        end=lb_types.Point(x=141.911, y=303.195),  # Bottom right
        page=1,
        unit=lb_types.RectangleUnit.POINTS))
bbox_labels = [
    lb_types.Label(data=lb_types.DocumentData(global_key='test-global-key'),
                   annotations=[bbox_annotation])
]
bbox_ndjson = [{
    'bbox': {
        'height': 216.697,
        'left': 42.799,
        'top': 86.498,
        'width': 99.112,
    },
    'classifications': [],
    'dataRow': {
        'globalKey': 'test-global-key'
    },
    'name': 'bounding_box',
    'page': 1,
    'unit': 'POINTS'
}]


def round_dict(data):
    if isinstance(data, dict):
        for key in data:
            if isinstance(data[key], (int, float)):
                data[key] = int(data[key])
            elif isinstance(data[key], dict):
                data[key] = round_dict(data[key])
            elif isinstance(data[key], (list, tuple)):
                data[key] = [round_dict(r) for r in data[key]]

    return data


def test_pdf():
    """
    Tests a pdf file with bbox annotations only
    """
    with open('tests/data/assets/ndjson/pdf_import.json', 'r') as f:
        data = json.load(f)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert [round_dict(x) for x in res] == [round_dict(x) for x in data]
    f.close()


def test_pdf_with_name_only():
    """
    Tests a pdf file with bbox annotations only
    """
    with open('tests/data/assets/ndjson/pdf_import_name_only.json', 'r') as f:
        data = json.load(f)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert [round_dict(x) for x in res] == [round_dict(x) for x in data]
    f.close()


def test_pdf_bbox_serialize():
    serialized = list(NDJsonConverter.serialize(bbox_labels))
    serialized[0].pop('uuid')
    assert serialized == bbox_ndjson


def test_pdf_bbox_deserialize():
    deserialized = list(NDJsonConverter.deserialize(bbox_ndjson))
    deserialized[0].annotations[0].extra = {}
    assert deserialized[0].annotations == bbox_labels[0].annotations



import json

from labelbox.data.serialization.labelbox_v1.converter import LBV1Converter


def test_text():
    with open('tests/data/assets/labelbox_v1/text_export.json', 'r') as file:
        payload = json.load(file)
    collection = LBV1Converter.deserialize([payload])
    serialized = next(LBV1Converter.serialize(collection))

    payload['media_type'] = 'text'
    payload['Global Key'] = None

    assert serialized.keys() == payload.keys()
    for key in serialized:
        if key != 'Label':
            assert serialized[key] == payload[key]
        elif key == 'Label':
            for annotation_a, annotation_b in zip(serialized[key]['objects'],
                                                  payload[key]['objects']):
                annotation_b['page'] = None
                annotation_b['unit'] = None
                if not len(annotation_a['classifications']):
                    # We don't add a classification key to the payload if there is no classifications.
                    annotation_a.pop('classifications')
                assert annotation_a == annotation_b



"""
This module imports types that differ across python versions, so other modules 
don't have to worry about where they should be imported from.
"""

import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


from labelbox.utils import _CamelCaseMixin

from labelbox import pydantic_compat

from typing import Any, Dict, Optional


class App(_CamelCaseMixin, pydantic_compat.BaseModel):
    id: Optional[str]
    model_id: str
    name: str
    description: Optional[str] = None
    inference_params: Dict[str, Any]
    class_to_schema_id: Dict[str, str]
    ontology_id: str
    created_by: Optional[str] = None

    @classmethod
    def type_name(cls):
        return "App"


APP_FIELD_NAMES = list(App.schema()['properties'].keys())



from typing import Optional


def pydantic_import(class_name, sub_module_path: Optional[str] = None):
    import importlib
    import importlib.metadata

    # Get the version of pydantic
    pydantic_version = importlib.metadata.version("pydantic")

    # Determine the module name based on the version
    module_name = "pydantic" if pydantic_version.startswith(
        "1") else "pydantic.v1"
    module_name = f"{module_name}.{sub_module_path}" if sub_module_path else module_name

    # Import the class from the module
    klass = getattr(importlib.import_module(module_name), class_name)

    return klass


BaseModel = pydantic_import("BaseModel")
PrivateAttr = pydantic_import("PrivateAttr")
Field = pydantic_import("Field")
ModelField = pydantic_import("ModelField", "fields")
ValidationError = pydantic_import("ValidationError")
ErrorWrapper = pydantic_import("ErrorWrapper", "error_wrappers")

validator = pydantic_import("validator")
root_validator = pydantic_import("root_validator")
conint = pydantic_import("conint")
conlist = pydantic_import("conlist")
constr = pydantic_import("constr")
confloat = pydantic_import("confloat")



import pytest

from labelbox.schema.annotation_import import AnnotationImport


def test_data_row_validation_errors():
    predictions = [
        {
            "answer": {
                "schemaId": "ckrb1sfl8099g0y91cxbd5ftb",
            },
            "schemaId": "c123",
            "dataRow": {
                "globalKey": "05e8ee85-072e-4eb2-b30a-501dee9b0d9d"
            },
        },
        {
            "answer": {
                "schemaId": "ckrb1sfl8099g0y91cxbd5ftb",
            },
            "schemaId": "c123",
            "dataRow": {
                "globalKey": "05e8ee85-072e-4eb2-b30a-501dee9b0d9d"
            },
        },
        {
            "answer": {
                "schemaId": "ckrb1sfl8099g0y91cxbd5ftb",
            },
            "schemaId": "c123",
            "dataRow": {
                "globalKey": "05e8ee85-072e-4eb2-b30a-501dee9b0d9d"
            },
        },
        {
            "answer": {
                "schemaId": "ckrb1sfl8099g0y91cxbd5ftb",
            },
            "schemaId": "c123",
            "dataRow": {
                "globalKey": "05e8ee85-072e-4eb2-b30a-501dee9b0d9d"
            },
        },
        {
            "answer": {
                "schemaId": "ckrb1sfl8099g0y91cxbd5ftb",
            },
            "schemaId": "c123",
            "dataRow": {
                "globalKey": "05e8ee85-072e-4eb2-b30a-501dee9b0d9d"
            },
        },
    ]

    # Set up data for validation errors
    # Invalid: Remove 'dataRow' part entirely
    del predictions[0]['dataRow']

    # Invalid: Set both id and globalKey
    predictions[1]['dataRow'] = {
        'id': 'some id',
        'globalKey': 'some global key'
    }

    # Invalid: Set both id and globalKey to None
    predictions[2]['dataRow'] = {'id': None, 'globalKey': None}

    # Valid
    predictions[3]['dataRow'] = {
        'id': 'some id',
    }

    # Valid
    predictions[4]['dataRow'] = {
        'globalKey': 'some global key',
    }

    with pytest.raises(ValueError) as exc_info:
        AnnotationImport._validate_data_rows(predictions)
    exception_str = str(exc_info.value)
    assert "Found 3 annotations with errors" in exception_str
    assert "'dataRow' is missing in" in exception_str
    assert "Must provide only one of 'id' or 'globalKey' for 'dataRow'" in exception_str
    assert "'dataRow': {'id': 'some id', 'globalKey': 'some global key'}" in exception_str
    assert "'dataRow': {'id': None, 'globalKey': None}" in exception_str



import uuid
import datetime
import time
import requests
import pytest

from labelbox.data.annotation_types.annotation import ObjectAnnotation
from labelbox.schema.annotation_import import LabelImport
from labelbox import Dataset, Project

IMAGE_URL = "https://storage.googleapis.com/lb-artifacts-testing-public/sdk_integration_test/potato.jpeg"


@pytest.mark.skip(reason="broken export v1 api, to be retired soon")
def test_export_annotations_nested_checklist(
        client, configured_project_with_complex_ontology,
        wait_for_data_row_processing):
    project, data_row = configured_project_with_complex_ontology
    data_row = wait_for_data_row_processing(client, data_row)
    ontology = project.ontology().normalized

    tool = ontology["tools"][0]

    nested_check = [
        subc for subc in tool["classifications"]
        if subc["name"] == "test-checklist-class"
    ][0]

    data = [{
        "uuid":
            str(uuid.uuid4()),
        "schemaId":
            tool['featureSchemaId'],
        "dataRow": {
            "id": data_row.uid
        },
        "bbox": {
            "top": 20,
            "left": 20,
            "height": 50,
            "width": 50
        },
        "classifications": [{
            "schemaId":
                nested_check["featureSchemaId"],
            "answers": [
                {
                    "schemaId": nested_check["options"][0]["featureSchemaId"]
                },
                {
                    "schemaId": nested_check["options"][1]["featureSchemaId"]
                },
            ]
        }]
    }]
    task = LabelImport.create_from_objects(client, project.uid,
                                           f'label-import-{uuid.uuid4()}', data)
    task.wait_until_done()
    labels = project.label_generator()
    object_annotation = [
        annot for annot in next(labels).annotations
        if isinstance(annot, ObjectAnnotation)
    ][0]

    nested_class_answers = object_annotation.classifications[0].value.answer
    assert len(nested_class_answers) == 2


@pytest.mark.skip(reason="broken export v1 api, to be retired soon")
def test_export_filtered_dates(client,
                               configured_project_with_complex_ontology):
    project, data_row = configured_project_with_complex_ontology
    ontology = project.ontology().normalized

    tool = ontology["tools"][0]

    data = [{
        "uuid": str(uuid.uuid4()),
        "schemaId": tool['featureSchemaId'],
        "dataRow": {
            "id": data_row.uid
        },
        "bbox": {
            "top": 20,
            "left": 20,
            "height": 50,
            "width": 50
        }
    }]

    task = LabelImport.create_from_objects(client, project.uid,
                                           f'label-import-{uuid.uuid4()}', data)
    task.wait_until_done()

    regular_export = project.export_labels(download=True)
    assert len(regular_export) == 1

    filtered_export = project.export_labels(download=True, start="2020-01-01")
    assert len(filtered_export) == 1

    filtered_export_with_time = project.export_labels(
        download=True, start="2020-01-01 00:00:01")
    assert len(filtered_export_with_time) == 1

    empty_export = project.export_labels(download=True,
                                         start="2020-01-01",
                                         end="2020-01-02")
    assert len(empty_export) == 0


@pytest.mark.skip(reason="broken export v1 api, to be retired soon")
def test_export_filtered_activity(client,
                                  configured_project_with_complex_ontology):
    project, data_row = configured_project_with_complex_ontology
    ontology = project.ontology().normalized

    tool = ontology["tools"][0]

    data = [{
        "uuid": str(uuid.uuid4()),
        "schemaId": tool['featureSchemaId'],
        "dataRow": {
            "id": data_row.uid
        },
        "bbox": {
            "top": 20,
            "left": 20,
            "height": 50,
            "width": 50
        }
    }]

    task = LabelImport.create_from_objects(client, project.uid,
                                           f'label-import-{uuid.uuid4()}', data)
    task.wait_until_done()

    regular_export = project.export_labels(download=True)
    assert len(regular_export) == 1

    filtered_export = project.export_labels(
        download=True,
        last_activity_start="2020-01-01",
        last_activity_end=(datetime.datetime.now() +
                           datetime.timedelta(days=2)).strftime("%Y-%m-%d"))
    assert len(filtered_export) == 1

    filtered_export_with_time = project.export_labels(
        download=True, last_activity_start="2020-01-01 00:00:01")
    assert len(filtered_export_with_time) == 1

    empty_export = project.export_labels(
        download=True,
        last_activity_start=(datetime.datetime.now() +
                             datetime.timedelta(days=2)).strftime("%Y-%m-%d"),
    )

    empty_export = project.export_labels(
        download=True,
        last_activity_end=(datetime.datetime.now() -
                           datetime.timedelta(days=1)).strftime("%Y-%m-%d"))
    assert len(empty_export) == 0


def test_export_data_rows(project: Project, dataset: Dataset):
    n_data_rows = 2
    task = dataset.create_data_rows([
        {
            "row_data": IMAGE_URL,
            "external_id": "my-image"
        },
    ] * n_data_rows)
    task.wait_till_done()

    data_rows = [dr.uid for dr in list(dataset.export_data_rows())]
    batch = project.create_batch("batch test", data_rows)
    result = list(batch.export_data_rows())
    exported_data_rows = [dr.uid for dr in result]

    assert len(result) == n_data_rows
    assert set(data_rows) == set(exported_data_rows)


def test_queued_data_row_export(configured_project):
    result = configured_project.export_queued_data_rows()
    assert len(result) == 1


@pytest.mark.skip(reason="broken export v1 api, to be retired soon")
def test_label_export(configured_project_with_label):
    project, _, _, label = configured_project_with_label
    label_id = label.uid
    # Wait for exporter to retrieve latest labels
    time.sleep(10)

    # TODO: Move to export_v2
    exported_labels_url = project.export_labels()
    assert exported_labels_url is not None
    exported_labels = requests.get(exported_labels_url)
    labels = [example['ID'] for example in exported_labels.json()]
    assert labels[0] == label_id
    #TODO: Add test for bulk export back.
    # The new exporter doesn't work with the create_label mutation


def test_issues_export(project):
    exported_issues_url = project.export_issues()
    assert exported_issues_url

    exported_issues_url = project.export_issues("Open")
    assert exported_issues_url
    assert "?status=Open" in exported_issues_url

    exported_issues_url = project.export_issues("Resolved")
    assert exported_issues_url
    assert "?status=Resolved" in exported_issues_url

    invalidStatusValue = "Closed"
    with pytest.raises(ValueError) as exc_info:
        exported_issues_url = project.export_issues(invalidStatusValue)
    assert "status must be in" in str(exc_info.value)
    assert "Found %s" % (invalidStatusValue) in str(exc_info.value)


def test_dataset_export(dataset, image_url):
    n_data_rows = 2
    ids = set()
    for _ in range(n_data_rows):
        ids.add(dataset.create_data_row(row_data=image_url))
    result = list(dataset.export_data_rows())
    assert len(result) == n_data_rows
    assert set(result) == ids


@pytest.mark.skip(reason="broken export v1 api, to be retired soon")
def test_data_row_export_with_empty_media_attributes(
        client, configured_project_with_label, wait_for_data_row_processing):
    project, _, data_row, _ = configured_project_with_label
    data_row = wait_for_data_row_processing(client, data_row)
    labels = list(project.label_generator())
    assert len(
        labels
    ) == 1, "Label export job unexpectedly returned an empty result set`"
    assert labels[0].data.media_attributes == {}



from labelbox.typing_imports import Literal
from labelbox.utils import _NoCoercionMixin
from .base_data import BaseData


class ConversationData(BaseData):
    class_name: Literal["ConversationData"] = "ConversationData"


from typing import Callable, Optional

import requests
from requests.exceptions import ConnectTimeout
from google.api_core import retry

from labelbox import pydantic_compat
from labelbox.exceptions import InternalServerError
from labelbox.typing_imports import Literal
from labelbox.utils import _NoCoercionMixin
from .base_data import BaseData


class TextData(BaseData, _NoCoercionMixin):
    """
    Represents text data. Requires arg file_path, text, or url

    >>> TextData(text="")

    Args:
        file_path (str)
        text (str)
        url (str)
    """
    class_name: Literal["TextData"] = "TextData"
    file_path: Optional[str] = None
    text: Optional[str] = None
    url: Optional[str] = None

    @property
    def value(self) -> str:
        """
        Property that unifies the data access pattern for all references to the text.

        Returns:
            string representation of the text
        """
        if self.text:
            return self.text
        elif self.file_path:
            with open(self.file_path, "r") as file:
                text = file.read()
            self.text = text
            return text
        elif self.url:
            text = self.fetch_remote()
            self.text = text
            return text
        else:
            raise ValueError("Must set either url, file_path or im_bytes")

    def set_fetch_fn(self, fn):
        object.__setattr__(self, 'fetch_remote', lambda: fn(self))

    @retry.Retry(deadline=15.,
                 predicate=retry.if_exception_type(ConnectTimeout,
                                                   InternalServerError))
    def fetch_remote(self) -> str:
        """
        Method for accessing url.

        If url is not publicly accessible or requires another access pattern
        simply override this function
        """
        response = requests.get(self.url)
        if response.status_code in [500, 502, 503, 504]:
            raise labelbox.exceptions.InternalServerError(response.text)
        response.raise_for_status()
        return response.text

    @retry.Retry(deadline=15.)
    def create_url(self, signer: Callable[[bytes], str]) -> None:
        """
        Utility for creating a url from any of the other text references.

        Args:
            signer: A function that accepts bytes and returns a signed url.
        Returns:
            url for the text
        """
        if self.url is not None:
            return self.url
        elif self.file_path is not None:
            with open(self.file_path, 'rb') as file:
                self.url = signer(file.read())
        elif self.text is not None:
            self.url = signer(self.text.encode())
        else:
            raise ValueError(
                "One of url, im_bytes, file_path, numpy must not be None.")
        return self.url

    @pydantic_compat.root_validator
    def validate_date(cls, values):
        file_path = values.get("file_path")
        text = values.get("text")
        url = values.get("url")
        uid = values.get('uid')
        global_key = values.get('global_key')
        if uid == file_path == text == url == global_key == None:
            raise ValueError(
                "One of `file_path`, `text`, `uid`, `global_key` or `url` required."
            )
        return values

    def __repr__(self) -> str:
        return  f"TextData(file_path={self.file_path}," \
                f"text={self.text[:30] + '...' if self.text is not None else None}," \
                f"url={self.url})"

    class config:
        # Required for discriminating between data types
        extra = 'forbid'



from typing import Dict, Optional, Union
from enum import Enum

from .base import ConfidenceValue, BaseMetric

from labelbox import pydantic_compat

ScalarMetricValue = pydantic_compat.confloat(ge=0, le=100_000_000)
ScalarMetricConfidenceValue = Dict[ConfidenceValue, ScalarMetricValue]


class ScalarMetricAggregation(Enum):
    ARITHMETIC_MEAN = "ARITHMETIC_MEAN"
    GEOMETRIC_MEAN = "GEOMETRIC_MEAN"
    HARMONIC_MEAN = "HARMONIC_MEAN"
    SUM = "SUM"


RESERVED_METRIC_NAMES = ('true_positive_count', 'false_positive_count',
                         'true_negative_count', 'false_negative_count',
                         'precision', 'recall', 'f1', 'iou')


class ScalarMetric(BaseMetric):
    """ Class representing scalar metrics

    For backwards compatibility, metric_name is optional.
    The metric_name will be set to a default name in the editor if it is not set.
    This is not recommended and support for empty metric_name fields will be removed.
    aggregation will be ignored wihtout providing a metric name.
    """
    metric_name: Optional[str] = None
    value: Union[ScalarMetricValue, ScalarMetricConfidenceValue]
    aggregation: ScalarMetricAggregation = ScalarMetricAggregation.ARITHMETIC_MEAN

    @pydantic_compat.validator('metric_name')
    def validate_metric_name(cls, name: Union[str, None]):
        if name is None:
            return None
        clean_name = name.lower().strip()
        if clean_name in RESERVED_METRIC_NAMES:
            raise ValueError(f"`{clean_name}` is a reserved metric name. "
                             "Please provide another value for `metric_name`.")
        return name

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        if res.get('metric_name') is None:
            res.pop('aggregation')
        return res



from labelbox import pydantic_compat
from enum import Enum
from labelbox.data.annotation_types.annotation import BaseAnnotation, ObjectAnnotation


class Relationship(pydantic_compat.BaseModel):

    class Type(Enum):
        UNIDIRECTIONAL = "unidirectional"
        BIDIRECTIONAL = "bidirectional"

    source: ObjectAnnotation
    target: ObjectAnnotation
    type: Type = Type.UNIDIRECTIONAL


class RelationshipAnnotation(BaseAnnotation):
    value: Relationship



import os

import pytest

from labelbox.data.annotation_types import TextData
from labelbox import pydantic_compat


def test_validate_schema():
    with pytest.raises(pydantic_compat.ValidationError):
        data = TextData()


def test_text():
    text = "hello world"
    metadata = []
    media_attributes = {}
    text_data = TextData(text=text,
                         metadata=metadata,
                         media_attributes=media_attributes)
    assert text_data.text == text


def test_url():
    url = "https://storage.googleapis.com/lb-artifacts-testing-public/sdk_integration_test/sample3.txt"
    text_data = TextData(url=url)
    text = text_data.value
    assert len(text) == 3541


def test_file(tmpdir):
    content = "foo bar baz"
    file = "hello.txt"
    dir = tmpdir.mkdir('data')
    dir.join(file).write(content)
    text_data = TextData(file_path=os.path.join(dir.strpath, file))
    assert len(text_data.value) == len(content)


def test_ref():
    external_id = "external_id"
    uid = "uid"
    metadata = []
    media_attributes = {}
    data = TextData(text="hello world",
                    external_id=external_id,
                    uid=uid,
                    metadata=metadata,
                    media_attributes=media_attributes)
    assert data.external_id == external_id
    assert data.uid == uid
    assert data.media_attributes == media_attributes
    assert data.metadata == metadata



from typing import Union
from labelbox import pydantic_compat
from .base import NDAnnotation, DataRow
from ...annotation_types.data import ImageData, TextData
from ...annotation_types.relationship import RelationshipAnnotation
from ...annotation_types.relationship import Relationship
from .objects import NDObjectType
from .base import DataRow

SUPPORTED_ANNOTATIONS = NDObjectType


class _Relationship(pydantic_compat.BaseModel):
    source: str
    target: str
    type: str


class NDRelationship(NDAnnotation):
    relationship: _Relationship

    @staticmethod
    def to_common(annotation: "NDRelationship", source: SUPPORTED_ANNOTATIONS,
                  target: SUPPORTED_ANNOTATIONS) -> RelationshipAnnotation:
        return RelationshipAnnotation(name=annotation.name,
                                      value=Relationship(
                                          source=source,
                                          target=target,
                                          type=Relationship.Type(
                                              annotation.relationship.type)),
                                      extra={'uuid': annotation.uuid},
                                      feature_schema_id=annotation.schema_id)

    @classmethod
    def from_common(cls, annotation: RelationshipAnnotation,
                    data: Union[ImageData, TextData]) -> "NDRelationship":
        relationship = annotation.value
        return cls(uuid=str(annotation._uuid),
                   name=annotation.name,
                   dataRow=DataRow(id=data.uid, global_key=data.global_key),
                   relationship=_Relationship(
                       source=str(relationship.source._uuid),
                       target=str(relationship.target._uuid),
                       type=relationship.type.value))



import json
import numpy as np
import cv2

from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.data.annotation_types import Mask, Label, ObjectAnnotation, ImageData, MaskData


def round_dict(data):
    if isinstance(data, dict):
        for key in data:
            if isinstance(data[key], float):
                data[key] = int(data[key])
            elif isinstance(data[key], dict):
                data[key] = round_dict(data[key])
            elif isinstance(data[key], (list, tuple)):
                data[key] = [round_dict(r) for r in data[key]]

    return data


def test_image():
    with open('tests/data/assets/ndjson/image_import.json', 'r') as file:
        data = json.load(file)

    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))

    for r in res:
        r.pop('classifications', None)
    assert [round_dict(x) for x in res] == [round_dict(x) for x in data]


def test_image_with_name_only():
    with open('tests/data/assets/ndjson/image_import_name_only.json',
              'r') as file:
        data = json.load(file)

    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    for r in res:
        r.pop('classifications', None)
    assert [round_dict(x) for x in res] == [round_dict(x) for x in data]


def test_mask():
    data = [{
        "uuid": "b862c586-8614-483c-b5e6-82810f70cac0",
        "schemaId": "ckrazcueb16og0z6609jj7y3y",
        "dataRow": {
            "id": "ckrazctum0z8a0ybc0b0o0g0v"
        },
        "mask": {
            "png":
                "iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAAAAACoWZBhAAAAMklEQVR4nD3MuQ3AQADDMOqQ/Vd2ijytaSiZLAcYuyLEYYYl9cvrlGftTHvsYl+u/3EDv0QLI8Z7FlwAAAAASUVORK5CYII="
        },
        "confidence": 0.8,
        "customMetrics": [{
            "name": "customMetric1",
            "value": 0.4
        }],
    }, {
        "uuid": "751fc725-f7b6-48ed-89b0-dd7d94d08af6",
        "schemaId": "ckrazcuec16ok0z66f956apb7",
        "dataRow": {
            "id": "ckrazctum0z8a0ybc0b0o0g0v"
        },
        "mask": {
            "instanceURI":
                "https://storage.labelbox.com/ckqcx1czn06830y61gh9v02cs%2F3e729327-f038-f66c-186e-45e921ef9717-1?Expires=1626806874672&KeyName=labelbox-assets-key-3&Signature=YsUOGKrsqmAZ68vT9BlPJOaRyLY",
            "colorRGB": [255, 0, 0]
        }
    }]
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    for r in res:
        r.pop('classifications', None)

    assert [round_dict(x) for x in res] == [round_dict(x) for x in data]


def test_mask_from_arr():
    mask_arr = np.round(np.zeros((32, 32))).astype(np.uint8)
    mask_arr = cv2.rectangle(mask_arr, (5, 5), (10, 10), (1, 1), -1)

    label = Label(annotations=[
        ObjectAnnotation(feature_schema_id="1" * 25,
                         value=Mask(mask=MaskData.from_2D_arr(arr=mask_arr),
                                    color=(1, 1, 1)))
    ],
                  data=ImageData(uid="0" * 25))
    res = next(NDJsonConverter.serialize([label]))
    res.pop("uuid")
    assert res == {
        "classifications": [],
        "schemaId": "1" * 25,
        "dataRow": {
            "id": "0" * 25
        },
        "mask": {
            "png":
                "iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAAAAABWESUoAAAAHklEQVR4nGNgGAKAEYn8j00BEyETBoOCUTAKhhwAAJW+AQwvpePVAAAAAElFTkSuQmCC"
        }
    }



import uuid

from labelbox import parser
from labelbox.schema.annotation_import import MALPredictionImport
"""
- Here we only want to check that the uploads are calling the validation
- Then with unit tests we can check the types of errors raised

"""


def test_create_with_url_arg(client, module_project,
                             annotation_import_test_helpers):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    label_import = MALPredictionImport.create(
        client=client,
        id=module_project.uid,
        name=name,
        url=url)
    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name, url)


def test_create_with_labels_arg(client, module_project, object_predictions,
                                annotation_import_test_helpers):
    """this test should check running state only to validate running, not completed"""
    name = str(uuid.uuid4())

    label_import = MALPredictionImport.create(client=client,
                                              id=module_project.uid,
                                              name=name,
                                              labels=object_predictions)

    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name)
    annotation_import_test_helpers.assert_file_content(
        label_import.input_file_url, object_predictions)


def test_create_with_path_arg(client, tmp_path, configured_project, object_predictions,
                              annotation_import_test_helpers):
    project = configured_project
    name = str(uuid.uuid4())
    file_name = f"{name}.ndjson"
    file_path = tmp_path / file_name
    with file_path.open("w") as f:
        parser.dump(object_predictions, f)

    label_import = MALPredictionImport.create(client=client,
                                              id=project.uid,
                                              name=name,
                                              path=str(file_path))

    assert label_import.parent_id == project.uid
    annotation_import_test_helpers.check_running_state(label_import, name)
    annotation_import_test_helpers.assert_file_content(
        label_import.input_file_url, object_predictions)



from typing import List

from labelbox import pydantic_compat
from labelbox.utils import _CamelCaseMixin


class DocumentTextSelection(_CamelCaseMixin, pydantic_compat.BaseModel):
    token_ids: List[str]
    group_id: str
    page: int

    @pydantic_compat.validator("page")
    def validate_page(cls, v):
        if v < 1:
            raise ValueError("Page must be greater than 1")
        return v


class DocumentEntity(_CamelCaseMixin, pydantic_compat.BaseModel):
    """ Represents a text entity """
    text_selections: List[DocumentTextSelection]



from unittest.mock import MagicMock

from labelbox.schema.export_task import Converter, JsonConverter, Range, _MetadataFileInfo


class TestJsonConverter:

    def test_with_correct_ndjson(self, generate_random_ndjson):
        line_count = 10
        ndjson = generate_random_ndjson(line_count)
        file_content = "\n".join(ndjson) + "\n"
        input_args = Converter.ConverterInputArgs(
            ctx=MagicMock(),
            file_info=_MetadataFileInfo(
                offsets=Range(start=0, end=len(file_content) - 1),
                lines=Range(start=0, end=line_count - 1),
                file="file.ndjson",
            ),
            raw_data=file_content,
        )
        with JsonConverter() as converter:
            current_offset = 0
            for idx, output in enumerate(converter.convert(input_args)):
                assert output.current_line == idx
                assert output.current_offset == current_offset
                assert output.json_str == ndjson[idx]
                current_offset += len(output.json_str) + 1

    def test_with_no_newline_at_end(self, generate_random_ndjson):
        line_count = 10
        ndjson = generate_random_ndjson(line_count)
        file_content = "\n".join(ndjson)
        input_args = Converter.ConverterInputArgs(
            ctx=MagicMock(),
            file_info=_MetadataFileInfo(
                offsets=Range(start=0, end=len(file_content) - 1),
                lines=Range(start=0, end=line_count - 1),
                file="file.ndjson",
            ),
            raw_data=file_content,
        )
        with JsonConverter() as converter:
            current_offset = 0
            for idx, output in enumerate(converter.convert(input_args)):
                assert output.current_line == idx
                assert output.current_offset == current_offset
                assert output.json_str == ndjson[idx]
                current_offset += len(output.json_str) + 1

    def test_from_offset(self, generate_random_ndjson):
        # testing middle of a JSON string, but not the last line
        line_count = 10
        line_start = 5
        ndjson = generate_random_ndjson(line_count)
        file_content = "\n".join(ndjson) + "\n"
        offset_end = len(file_content)
        skipped_bytes = 15
        current_offset = file_content.find(ndjson[line_start]) + skipped_bytes
        file_content = file_content[current_offset:]

        input_args = Converter.ConverterInputArgs(
            ctx=MagicMock(),
            file_info=_MetadataFileInfo(
                offsets=Range(start=current_offset, end=offset_end),
                lines=Range(start=line_start, end=line_count - 1),
                file="file.ndjson",
            ),
            raw_data=file_content,
        )
        with JsonConverter() as converter:
            for idx, output in enumerate(converter.convert(input_args)):
                assert output.current_line == line_start + idx
                assert output.current_offset == current_offset
                assert output.json_str == ndjson[line_start +
                                                 idx][skipped_bytes:]
                current_offset += len(output.json_str) + 1
                skipped_bytes = 0

    def test_from_offset_last_line(self, generate_random_ndjson):
        # testing middle of a JSON string, but not the last line
        line_count = 10
        line_start = 9
        ndjson = generate_random_ndjson(line_count)
        file_content = "\n".join(ndjson) + "\n"
        offset_end = len(file_content)
        skipped_bytes = 15
        current_offset = file_content.find(ndjson[line_start]) + skipped_bytes
        file_content = file_content[current_offset:]

        input_args = Converter.ConverterInputArgs(
            ctx=MagicMock(),
            file_info=_MetadataFileInfo(
                offsets=Range(start=current_offset, end=offset_end),
                lines=Range(start=line_start, end=line_count - 1),
                file="file.ndjson",
            ),
            raw_data=file_content,
        )
        with JsonConverter() as converter:
            for idx, output in enumerate(converter.convert(input_args)):
                assert output.current_line == line_start + idx
                assert output.current_offset == current_offset
                assert output.json_str == ndjson[line_start +
                                                 idx][skipped_bytes:]
                current_offset += len(output.json_str) + 1
                skipped_bytes = 0



from typing import Any, Dict, List, Optional, Union, Type
try:
    from typing import Literal
except:
    from typing_extensions import Literal

from labelbox import pydantic_compat
import numpy as np

from .classification import LBV1Checklist, LBV1Classifications, LBV1Radio, LBV1Text, LBV1Dropdown
from .feature import LBV1Feature
from ...annotation_types.annotation import (ClassificationAnnotation,
                                            ObjectAnnotation)
from ...annotation_types.data import MaskData
from ...annotation_types.geometry import Line, Mask, Point, Polygon, Rectangle
from ...annotation_types.ner import TextEntity
from ...annotation_types.types import Cuid


class LBV1ObjectBase(LBV1Feature):
    color: Optional[str] = None
    instanceURI: Optional[str] = None
    classifications: List[Union[LBV1Text, LBV1Radio, LBV1Dropdown,
                                LBV1Checklist]] = []
    page: Optional[int] = None
    unit: Optional[str] = None

    def dict(self, *args, **kwargs) -> Dict[str, Any]:
        res = super().dict(*args, **kwargs)
        # This means these are not video frames ..
        if self.instanceURI is None:
            res.pop('instanceURI')
        return res

    @pydantic_compat.validator('classifications', pre=True)
    def validate_subclasses(cls, value, field):
        # checklist subclasses create extra unessesary nesting. So we just remove it.
        if isinstance(value, list) and len(value):
            subclasses = []
            for v in value:
                # this is due to Checklists providing extra brackets []. We grab every item
                # in the brackets if this is the case
                if isinstance(v, list):
                    for inner_v in v:
                        subclasses.append(inner_v)
                else:
                    subclasses.append(v)
            return subclasses
        return value


class TIPointCoordinate(pydantic_compat.BaseModel):
    coordinates: List[float]


class TILineCoordinate(pydantic_compat.BaseModel):
    coordinates: List[List[float]]


class TIPolygonCoordinate(pydantic_compat.BaseModel):
    coordinates: List[List[List[float]]]


class TIRectangleCoordinate(pydantic_compat.BaseModel):
    coordinates: List[List[List[float]]]


class LBV1TIPoint(LBV1ObjectBase):
    object_type: Literal['point'] = pydantic_compat.Field(..., alias='type')
    geometry: TIPointCoordinate

    def to_common(self) -> Point:
        lng, lat = self.geometry.coordinates
        return Point(x=lng, y=lat)


class LBV1TILine(LBV1ObjectBase):
    object_type: Literal['polyline'] = pydantic_compat.Field(..., alias='type')
    geometry: TILineCoordinate

    def to_common(self) -> Line:
        return Line(points=[
            Point(x=coord[0], y=coord[1]) for coord in self.geometry.coordinates
        ])


class LBV1TIPolygon(LBV1ObjectBase):
    object_type: Literal['polygon'] = pydantic_compat.Field(..., alias='type')
    geometry: TIPolygonCoordinate

    def to_common(self) -> Polygon:
        for coord_list in self.geometry.coordinates:
            return Polygon(
                points=[Point(x=coord[0], y=coord[1]) for coord in coord_list])


class LBV1TIRectangle(LBV1ObjectBase):
    object_type: Literal['rectangle'] = pydantic_compat.Field(..., alias='type')
    geometry: TIRectangleCoordinate

    def to_common(self) -> Rectangle:
        coord_list = np.array(self.geometry.coordinates[0])

        min_x, max_x = np.min(coord_list[:, 0]), np.max(coord_list[:, 0])
        min_y, max_y = np.min(coord_list[:, 1]), np.max(coord_list[:, 1])

        start = [min_x, min_y]
        end = [max_x, max_y]

        return Rectangle(start=Point(x=start[0], y=start[1]),
                         end=Point(x=end[0], y=end[1]))


class _Point(pydantic_compat.BaseModel):
    x: float
    y: float


class _Box(pydantic_compat.BaseModel):
    top: float
    left: float
    height: float
    width: float


class LBV1Rectangle(LBV1ObjectBase):
    bbox: _Box

    def to_common(self) -> Rectangle:
        return Rectangle(start=Point(x=self.bbox.left, y=self.bbox.top),
                         end=Point(x=self.bbox.left + self.bbox.width,
                                   y=self.bbox.top + self.bbox.height))

    @classmethod
    def from_common(cls, rectangle: Rectangle,
                    classifications: List[ClassificationAnnotation],
                    feature_schema_id: Cuid, title: str,
                    extra: Dict[str, Any]) -> "LBV1Rectangle":
        return cls(bbox=_Box(
            top=rectangle.start.y,
            left=rectangle.start.x,
            height=rectangle.end.y - rectangle.start.y,
            width=rectangle.end.x - rectangle.start.x,
        ),
                   schema_id=feature_schema_id,
                   title=title,
                   classifications=classifications,
                   **extra)


class LBV1Polygon(LBV1ObjectBase):
    polygon: List[_Point]

    def to_common(self) -> Polygon:
        return Polygon(points=[Point(x=p.x, y=p.y) for p in self.polygon])

    @classmethod
    def from_common(cls, polygon: Polygon,
                    classifications: List[ClassificationAnnotation],
                    feature_schema_id: Cuid, title: str,
                    extra: Dict[str, Any]) -> "LBV1Polygon":
        return cls(
            polygon=[
                _Point(x=point.x, y=point.y) for point in polygon.points[:-1]
            ],  # drop closing point
            classifications=classifications,
            schema_id=feature_schema_id,
            title=title,
            **extra)


class LBV1Point(LBV1ObjectBase):
    point: _Point

    def to_common(self) -> Point:
        return Point(x=self.point.x, y=self.point.y)

    @classmethod
    def from_common(cls, point: Point,
                    classifications: List[ClassificationAnnotation],
                    feature_schema_id: Cuid, title: str,
                    extra: Dict[str, Any]) -> "LBV1Point":
        return cls(point=_Point(x=point.x, y=point.y),
                   classifications=classifications,
                   schema_id=feature_schema_id,
                   title=title,
                   **extra)


class LBV1Line(LBV1ObjectBase):
    line: List[_Point]

    def to_common(self) -> Line:
        return Line(points=[Point(x=p.x, y=p.y) for p in self.line])

    @classmethod
    def from_common(cls, polygon: Line,
                    classifications: List[ClassificationAnnotation],
                    feature_schema_id: Cuid, title: str,
                    extra: Dict[str, Any]) -> "LBV1Line":
        return cls(
            line=[_Point(x=point.x, y=point.y) for point in polygon.points],
            classifications=classifications,
            schema_id=feature_schema_id,
            title=title,
            **extra)


class LBV1Mask(LBV1ObjectBase):
    instanceURI: str

    def to_common(self) -> Mask:
        return Mask(mask=MaskData(url=self.instanceURI), color=(255, 255, 255))

    @classmethod
    def from_common(cls, mask: Mask,
                    classifications: List[ClassificationAnnotation],
                    feature_schema_id: Cuid, title: str,
                    extra: Dict[str, Any]) -> "LBV1Mask":
        if mask.mask.url is None:
            raise ValueError(
                "Mask does not have a url. Use `LabelGenerator.add_url_to_masks`, or `Label.add_url_to_masks`."
            )
        return cls(instanceURI=mask.mask.url,
                   classifications=classifications,
                   schema_id=feature_schema_id,
                   title=title,
                   **{
                       k: v for k, v in extra.items() if k != 'instanceURI'
                   })


class _TextPoint(pydantic_compat.BaseModel):
    start: int
    end: int


class _Location(pydantic_compat.BaseModel):
    location: _TextPoint


class LBV1TextEntity(LBV1ObjectBase):
    data: _Location
    format: str = "text.location"
    version: int = 1

    def to_common(self) -> TextEntity:
        return TextEntity(
            start=self.data.location.start,
            end=self.data.location.end,
        )

    @classmethod
    def from_common(cls, text_entity: TextEntity,
                    classifications: List[ClassificationAnnotation],
                    feature_schema_id: Cuid, title: str,
                    extra: Dict[str, Any]) -> "LBV1TextEntity":
        return cls(data=_Location(
            location=_TextPoint(start=text_entity.start, end=text_entity.end)),
                   classifications=classifications,
                   schema_id=feature_schema_id,
                   title=title,
                   **extra)


class LBV1Objects(pydantic_compat.BaseModel):
    objects: List[Union[
        LBV1Line,
        LBV1Point,
        LBV1Polygon,
        LBV1Rectangle,
        LBV1TextEntity,
        LBV1Mask,
        LBV1TIPoint,
        LBV1TILine,
        LBV1TIPolygon,
        LBV1TIRectangle,
    ]]

    def to_common(self) -> List[ObjectAnnotation]:
        objects = [
            ObjectAnnotation(value=obj.to_common(),
                             classifications=[
                                 ClassificationAnnotation(
                                     value=cls.to_common(),
                                     feature_schema_id=cls.schema_id,
                                     name=cls.title,
                                     extra={
                                         'feature_id': cls.feature_id,
                                         'title': cls.title,
                                         'value': cls.value
                                     }) for cls in obj.classifications
                             ],
                             name=obj.title,
                             feature_schema_id=obj.schema_id,
                             extra={
                                 'instanceURI': obj.instanceURI,
                                 'color': obj.color,
                                 'feature_id': obj.feature_id,
                                 'value': obj.value,
                                 'page': obj.page,
                                 'unit': obj.unit,
                             }) for obj in self.objects
        ]
        return objects

    @classmethod
    def from_common(cls, annotations: List[ObjectAnnotation]) -> "LBV1Objects":
        objects = []
        for annotation in annotations:
            obj = cls.lookup_object(annotation)
            subclasses = LBV1Classifications.from_common(
                annotation.classifications).classifications

            objects.append(
                obj.from_common(
                    annotation.value, subclasses, annotation.feature_schema_id,
                    annotation.name, {
                        'keyframe': getattr(annotation, 'keyframe', None),
                        **annotation.extra
                    }))
        return cls(objects=objects)

    @staticmethod
    def lookup_object(
        annotation: ObjectAnnotation
    ) -> Type[Union[LBV1Line, LBV1Point, LBV1Polygon, LBV1Rectangle, LBV1Mask,
                    LBV1TextEntity]]:
        result = {
            Line: LBV1Line,
            Point: LBV1Point,
            Polygon: LBV1Polygon,
            Rectangle: LBV1Rectangle,
            Mask: LBV1Mask,
            TextEntity: LBV1TextEntity
        }.get(type(annotation.value))
        if result is None:
            raise TypeError(f"Unexpected type {type(annotation.value)}")
        return result



from typing import Any, Dict, List, Optional, Tuple, Union
from labelbox.orm.db_object import experimental
from labelbox.schema.export_filters import CatalogExportFilters, build_filters

from labelbox.schema.export_params import (CatalogExportParams,
                                           validate_catalog_export_params)
from labelbox.schema.export_task import ExportTask
from labelbox.schema.task import Task

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from labelbox import Client


class Catalog:
    client: "Client"

    def __init__(self, client: 'Client'):
        self.client = client

    def export_v2(
        self,
        task_name: Optional[str] = None,
        filters: Union[CatalogExportFilters, Dict[str, List[str]], None] = None,
        params: Optional[CatalogExportParams] = None,
    ) -> Union[Task, ExportTask]:
        """
        Creates a catalog export task with the given params, filters and returns the task.

        >>>     import labelbox as lb
        >>>     client = lb.Client(<API_KEY>)
        >>>     catalog = client.get_catalog()
        >>>     task = catalog.export_v2(
        >>>         filters={
        >>>             "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>             "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>         },
        >>>         params={
        >>>             "performance_details": False,
        >>>             "label_details": True
        >>>         })
        >>>     task.wait_till_done()
        >>>     task.result
        """
        task, is_streamable = self._export(task_name, filters, params)
        if (is_streamable):
            return ExportTask(task, True)
        return task

    def export(
        self,
        task_name: Optional[str] = None,
        filters: Union[CatalogExportFilters, Dict[str, List[str]], None] = None,
        params: Optional[CatalogExportParams] = None,
    ) -> ExportTask:
        """
        Creates a catalog export task with the given params, filters and returns the task.

        >>>     import labelbox as lb
        >>>     client = lb.Client(<API_KEY>)
        >>>     export_task = Catalog.export(
        >>>         filters={
        >>>             "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>             "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>         },
        >>>         params={
        >>>             "performance_details": False,
        >>>             "label_details": True
        >>>         })
        >>>     export_task.wait_till_done()
        """
        task, _ = self._export(task_name, filters, params, streamable=True)
        return ExportTask(task)

    def _export(self,
                task_name: Optional[str] = None,
                filters: Union[CatalogExportFilters, Dict[str, List[str]],
                               None] = None,
                params: Optional[CatalogExportParams] = None,
                streamable: bool = False) -> Tuple[Task, bool]:

        _params = params or CatalogExportParams({
            "attachments": False,
            "embeddings": False,
            "metadata_fields": False,
            "data_row_details": False,
            "project_details": False,
            "performance_details": False,
            "label_details": False,
            "media_type_override": None,
            "model_run_ids": None,
            "project_ids": None,
            "interpolated_frames": False,
            "all_projects": False,
            "all_model_runs": False,
        })
        validate_catalog_export_params(_params)

        _filters = filters or CatalogExportFilters({
            "last_activity_at": None,
            "label_created_at": None,
            "data_row_ids": None,
            "global_keys": None,
        })

        mutation_name = "exportDataRowsInCatalog"
        create_task_query_str = (
            f"mutation {mutation_name}PyApi"
            f"($input: ExportDataRowsInCatalogInput!)"
            f"{{{mutation_name}(input: $input){{taskId isStreamable}}}}")

        media_type_override = _params.get('media_type_override', None)
        query_params: Dict[str, Any] = {
            "input": {
                "taskName": task_name,
                "filters": {
                    "searchQuery": {
                        "scope": None,
                        "query": None,
                    }
                },
                "isStreamableReady": True,
                "params": {
                    "mediaTypeOverride":
                        media_type_override.value
                        if media_type_override is not None else None,
                    "includeAttachments":
                        _params.get('attachments', False),
                    "includeEmbeddings":
                        _params.get('embeddings', False),
                    "includeMetadata":
                        _params.get('metadata_fields', False),
                    "includeDataRowDetails":
                        _params.get('data_row_details', False),
                    "includeProjectDetails":
                        _params.get('project_details', False),
                    "includePerformanceDetails":
                        _params.get('performance_details', False),
                    "includeLabelDetails":
                        _params.get('label_details', False),
                    "includeInterpolatedFrames":
                        _params.get('interpolated_frames', False),
                    "includePredictions":
                        _params.get('predictions', False),
                    "projectIds":
                        _params.get('project_ids', None),
                    "modelRunIds":
                        _params.get('model_run_ids', None),
                    "allProjects":
                        _params.get('all_projects', False),
                    "allModelRuns":
                        _params.get('all_model_runs', False),
                },
                "streamable": streamable,
            }
        }

        search_query = build_filters(self.client, _filters)
        query_params["input"]["filters"]["searchQuery"]["query"] = search_query

        res = self.client.execute(create_task_query_str,
                                  query_params,
                                  error_log_key="errors")
        res = res[mutation_name]
        task_id = res["taskId"]
        is_streamable = res["isStreamable"]
        return Task.get_task(self.client, task_id), is_streamable



import pytest

from labelbox import Tool, MediaType

point = Tool(
    tool=Tool.Type.POINT,
    name="name",
    color="#ff0000",
)


def test_deletes_a_feature_schema(client):
    tool = client.upsert_feature_schema(point.asdict())

    assert client.delete_unused_feature_schema(
        tool.normalized['featureSchemaId']) is None


def test_cant_delete_already_deleted_feature_schema(client):
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized['featureSchemaId']

    client.delete_unused_feature_schema(feature_schema_id) is None

    with pytest.raises(
            Exception,
            match=
            "Failed to delete the feature schema, message: Feature schema is already deleted"
    ):
        client.delete_unused_feature_schema(feature_schema_id)


def test_cant_delete_feature_schema_with_ontology(client):
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized['featureSchemaId']
    ontology = client.create_ontology_from_feature_schemas(
        name='ontology name',
        feature_schema_ids=[feature_schema_id],
        media_type=MediaType.Image)

    with pytest.raises(
            Exception,
            match=
            "Failed to delete the feature schema, message: Feature schema cannot be deleted because it is used in ontologies"
    ):
        client.delete_unused_feature_schema(feature_schema_id)

    client.delete_unused_ontology(ontology.uid)
    client.delete_unused_feature_schema(feature_schema_id)


def test_throws_an_error_if_feature_schema_to_delete_doesnt_exist(client):
    with pytest.raises(
            Exception,
            match=
            "Failed to delete the feature schema, message: Cannot find root schema node with feature schema id doesntexist"
    ):
        client.delete_unused_feature_schema("doesntexist")


def test_updates_a_feature_schema_title(client):
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized['featureSchemaId']
    new_title = "new title"
    updated_feature_schema = client.update_feature_schema_title(
        feature_schema_id, new_title)

    assert updated_feature_schema.normalized['name'] == new_title

    client.delete_unused_feature_schema(feature_schema_id)


def test_throws_an_error_when_updating_a_feature_schema_with_empty_title(
        client):
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized['featureSchemaId']

    with pytest.raises(Exception):
        client.update_feature_schema_title(feature_schema_id, "")

    client.delete_unused_feature_schema(feature_schema_id)


def test_throws_an_error_when_updating_not_existing_feature_schema(client):
    with pytest.raises(Exception):
        client.update_feature_schema_title("doesntexist", "new title")


def test_creates_a_new_feature_schema(client):
    created_feature_schema = client.upsert_feature_schema(point.asdict())

    assert created_feature_schema.uid is not None

    client.delete_unused_feature_schema(
        created_feature_schema.normalized['featureSchemaId'])


def test_updates_a_feature_schema(client):
    tool = Tool(
        tool=Tool.Type.POINT,
        name="name",
        color="#ff0000",
    )
    created_feature_schema = client.upsert_feature_schema(tool.asdict())
    tool_to_update = Tool(
        tool=Tool.Type.POINT,
        name="new name",
        color="#ff0000",
        feature_schema_id=created_feature_schema.normalized['featureSchemaId'],
    )
    updated_feature_schema = client.upsert_feature_schema(
        tool_to_update.asdict())

    assert updated_feature_schema.normalized['name'] == "new name"


def test_does_not_include_used_feature_schema(client):
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized['featureSchemaId']
    ontology = client.create_ontology_from_feature_schemas(
        name='ontology name',
        feature_schema_ids=[feature_schema_id],
        media_type=MediaType.Image)
    unused_feature_schemas = client.get_unused_feature_schemas()

    assert feature_schema_id not in unused_feature_schemas

    client.delete_unused_ontology(ontology.uid)
    client.delete_unused_feature_schema(feature_schema_id)



from labelbox.typing_imports import Literal
from labelbox.utils import _NoCoercionMixin
from .base_data import BaseData


class DocumentData(BaseData, _NoCoercionMixin):
    class_name: Literal["DocumentData"] = "DocumentData"


from typing import List, Union

from labelbox import pydantic_compat

from .feature import LBV1Feature
from ...annotation_types.annotation import ClassificationAnnotation
from ...annotation_types.classification import Checklist, ClassificationAnswer, Radio, Text, Dropdown
from ...annotation_types.types import Cuid


class LBV1ClassificationAnswer(LBV1Feature):

    def to_common(self) -> ClassificationAnswer:
        return ClassificationAnswer(feature_schema_id=self.schema_id,
                                    name=self.title,
                                    keyframe=self.keyframe,
                                    extra={
                                        'feature_id': self.feature_id,
                                        'value': self.value
                                    })

    @classmethod
    def from_common(
            cls,
            answer: ClassificationAnnotation) -> "LBV1ClassificationAnswer":
        return cls(schema_id=answer.feature_schema_id,
                   title=answer.name,
                   value=answer.extra.get('value'),
                   feature_id=answer.extra.get('feature_id'),
                   keyframe=answer.keyframe)


class LBV1Radio(LBV1Feature):
    answer: LBV1ClassificationAnswer

    def to_common(self) -> Radio:
        return Radio(answer=self.answer.to_common())

    @classmethod
    def from_common(cls, radio: Radio, feature_schema_id: Cuid,
                    **extra) -> "LBV1Radio":
        return cls(schema_id=feature_schema_id,
                   answer=LBV1ClassificationAnswer.from_common(radio.answer),
                   **extra)


class LBV1Checklist(LBV1Feature):
    answers: List[LBV1ClassificationAnswer]

    def to_common(self) -> Checklist:
        return Checklist(answer=[answer.to_common() for answer in self.answers])

    @classmethod
    def from_common(cls, checklist: Checklist, feature_schema_id: Cuid,
                    **extra) -> "LBV1Checklist":
        return cls(schema_id=feature_schema_id,
                   answers=[
                       LBV1ClassificationAnswer.from_common(answer)
                       for answer in checklist.answer
                   ],
                   **extra)


class LBV1Dropdown(LBV1Feature):
    answer: List[LBV1ClassificationAnswer]

    def to_common(self) -> Dropdown:
        return Dropdown(answer=[answer.to_common() for answer in self.answer])

    @classmethod
    def from_common(cls, dropdown: Dropdown, feature_schema_id: Cuid,
                    **extra) -> "LBV1Dropdown":
        return cls(schema_id=feature_schema_id,
                   answer=[
                       LBV1ClassificationAnswer.from_common(answer)
                       for answer in dropdown.answer
                   ],
                   **extra)


class LBV1Text(LBV1Feature):
    answer: str

    def to_common(self) -> Text:
        return Text(answer=self.answer)

    @classmethod
    def from_common(cls, text: Text, feature_schema_id: Cuid,
                    **extra) -> "LBV1Text":
        return cls(schema_id=feature_schema_id, answer=text.answer, **extra)


class LBV1Classifications(pydantic_compat.BaseModel):
    classifications: List[Union[LBV1Text, LBV1Radio, LBV1Dropdown,
                                LBV1Checklist]] = []

    def to_common(self) -> List[ClassificationAnnotation]:
        classifications = [
            ClassificationAnnotation(value=classification.to_common(),
                                     name=classification.title,
                                     feature_schema_id=classification.schema_id,
                                     extra={
                                         'value': classification.value,
                                         'feature_id': classification.feature_id
                                     })
            for classification in self.classifications
        ]
        return classifications

    @classmethod
    def from_common(
            cls, annotations: List[ClassificationAnnotation]
    ) -> "LBV1Classifications":
        classifications = []
        for annotation in annotations:
            classification = cls.lookup_classification(annotation)
            if classification is not None:
                classifications.append(
                    classification.from_common(annotation.value,
                                               annotation.feature_schema_id,
                                               **annotation.extra))
            else:
                raise TypeError(f"Unexpected type {type(annotation.value)}")
        return cls(classifications=classifications)

    @staticmethod
    def lookup_classification(
        annotation: ClassificationAnnotation
    ) -> Union[LBV1Text, LBV1Checklist, LBV1Radio, LBV1Checklist]:
        return {
            Text: LBV1Text,
            Dropdown: LBV1Dropdown,
            Checklist: LBV1Checklist,
            Radio: LBV1Radio
        }.get(type(annotation.value))



import json
from labelbox.data.annotation_types.annotation import ClassificationAnnotation
from labelbox.data.annotation_types.classification.classification import ClassificationAnswer
from labelbox.data.annotation_types.classification.classification import Radio
from labelbox.data.annotation_types.data.text import TextData
from labelbox.data.annotation_types.label import Label

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_serialization_with_radio_min():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=TextData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
            text="This is a test",
        ),
        annotations=[
            ClassificationAnnotation(
                name="radio_question_geo",
                value=Radio(
                    answer=ClassificationAnswer(name="first_radio_answer",)))
        ])

    expected = {
        'name': 'radio_question_geo',
        'answer': {
            'name': 'first_radio_answer'
        },
        'dataRow': {
            'id': 'bkj7z2q0b0000jx6x0q2q7q0d'
        }
    }
    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)

    res.pop("uuid")
    assert res == expected

    deserialized = NDJsonConverter.deserialize([res])
    res = next(deserialized)
    res.annotations[0].extra.pop("uuid")
    assert res.annotations == label.annotations


def test_serialization_with_radio_classification():
    label = Label(uid="ckj7z2q0b0000jx6x0q2q7q0d",
                  data=TextData(
                      uid="bkj7z2q0b0000jx6x0q2q7q0d",
                      text="This is a test",
                  ),
                  annotations=[
                      ClassificationAnnotation(
                          name="radio_question_geo",
                          confidence=0.5,
                          value=Radio(answer=ClassificationAnswer(
                              confidence=0.6,
                              name="first_radio_answer",
                              classifications=[
                                  ClassificationAnnotation(
                                      name="sub_radio_question",
                                      value=Radio(answer=ClassificationAnswer(
                                          name="first_sub_radio_answer")))
                              ])))
                  ])

    expected = {
        'confidence': 0.5,
        'name': 'radio_question_geo',
        'answer': {
            'confidence':
                0.6,
            'name':
                'first_radio_answer',
            'classifications': [{
                'name': 'sub_radio_question',
                'answer': {
                    'name': 'first_sub_radio_answer',
                }
            }]
        },
        'dataRow': {
            'id': 'bkj7z2q0b0000jx6x0q2q7q0d'
        }
    }

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)
    res.pop("uuid")
    assert res == expected

    deserialized = NDJsonConverter.deserialize([res])
    res = next(deserialized)
    res.annotations[0].extra.pop("uuid")
    assert res.annotations == label.annotations



import json

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_classification():
    with open('tests/data/assets/ndjson/classification_import.json',
              'r') as file:
        data = json.load(file)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == data


def test_classification_with_name():
    with open('tests/data/assets/ndjson/classification_import_name_only.json',
              'r') as file:
        data = json.load(file)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == data
